# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" ULPI interfacing hardware. """

from amaranth       import Signal, Module, Cat, Elaboratable, ClockSignal, \
                           Record, ResetSignal, Const
from amaranth.hdl.rec import Record, DIR_FANIN, DIR_FANOUT, DIR_NONE

from ..utils.io     import delay


class ULPIInterface(Record):
    """ Record that represents a standard ULPI interface. """

    LAYOUT = [
        ('data', [('i', 8, DIR_FANIN), ('o', 8, DIR_FANOUT), ('oe', 1, DIR_FANOUT)]),
        ('clk', 1, DIR_FANOUT),
        ('nxt', 1, DIR_FANIN),
        ('stp', 1, DIR_FANOUT),
        ('dir', [('i', 1, DIR_FANIN)]),
        ('rst', 1, DIR_FANOUT)
    ]

    def __init__(self):
        super().__init__(self.LAYOUT)



class ULPIRegisterWindow(Elaboratable):
    """ Gateware interface that handles ULPI register reads and writes.

    I/O ports:

        # ULPI signals:
        I: ulpi_data_in[8]   -- input value of the ULPI data lines
        O: ulpi_data_out[8]  -- output value of the ULPI data lines
        O: ulpi_out_en       -- true iff we're trying to drive the ULPI data lines

        # Controller signals:
        O: busy              -- indicates when the register window is busy processing a transaction
        I: address[6]        -- the address of the register to work with
        O: done              -- strobe that indicates when a register request is complete

        I: read_request      -- strobe that requests a register read
        O: read_data[8]      -- data read from the relevant register read

        I: write_request     -- strobe that indicates a register write
        I: write_data[8]     -- data to be written during a register write

    """

    COMMAND_REG_WRITE = 0b10000000
    COMMAND_REG_READ  = 0b11000000

    def __init__(self):

        #
        # I/O port.
        #

        self.ulpi_data_in  = Signal(8)
        self.ulpi_data_out = Signal(8)
        self.ulpi_out_req  = Signal()
        self.ulpi_dir      = Signal()
        self.ulpi_next     = Signal()
        self.ulpi_stop     = Signal()

        self.busy          = Signal()
        self.address       = Signal(6)
        self.done          = Signal()

        self.read_request  = Signal()
        self.read_data     = Signal(8)

        self.write_request = Signal()
        self.write_data    = Signal(8)


    def elaborate(self, platform):
        m = Module()

        current_address = Signal(6)
        current_write   = Signal(8)

        # Keep our control signals low unless explicitly asserted.
        m.d.usb += [
            self.ulpi_out_req.eq(0),
            self.ulpi_stop   .eq(0),
            self.done        .eq(0)
        ]

        with m.FSM(domain="usb") as fsm:

            # We're busy whenever we're not IDLE; indicate so.
            m.d.comb += self.busy.eq(~fsm.ongoing('IDLE'))

            # IDLE: wait for a request to be made
            with m.State('IDLE'):

                # Apply a NOP whenever we're idle.
                #
                # This doesn't technically help for normal ULPI
                # operation, as the controller should handle this,
                # but it cleans up the output in our tests and allows
                # this unit to be used standalone.
                m.d.usb += self.ulpi_data_out.eq(0)

                # Constantly latch in our arguments while IDLE.
                # We'll stop latching these in as soon as we're busy.
                m.d.usb += [
                    current_address .eq(self.address),
                    current_write   .eq(self.write_data)
                ]

                with m.If(self.read_request):
                    m.next = 'START_READ'

                with m.If(self.write_request):
                    m.next = 'START_WRITE'

            #
            # Read handling.
            #

            # START_READ: wait for the bus to be idle, so we can transmit.
            with m.State('START_READ'):

                # Wait for the bus to be idle.
                with m.If(~self.ulpi_dir):
                    m.next = 'SEND_READ_ADDRESS'

                    # Once it is, start sending our command.
                    m.d.usb += [
                        self.ulpi_data_out .eq(self.COMMAND_REG_READ | self.address),
                        self.ulpi_out_req  .eq(1)
                    ]


            # SEND_READ_ADDRESS: Request sending the read address, which we
            # start sending on the next clock cycle. Note that we don't want
            # to come into this state writing, as we need to lead with a
            # bus-turnaround cycle.
            with m.State('SEND_READ_ADDRESS'):
                m.d.usb += self.ulpi_out_req.eq(1)

                # If DIR has become asserted, we're being interrupted.
                # We'll have to restart the read after the interruption is over.
                with m.If(self.ulpi_dir):
                    m.next = 'START_READ'
                    m.d.usb += self.ulpi_out_req.eq(0)

                # If NXT becomes asserted without us being interrupted by
                # DIR, then the PHY has accepted the read. Release our write
                # request, so the next cycle can properly act as a bus turnaround.
                with m.Elif(self.ulpi_next):
                    m.d.usb += [
                        self.ulpi_out_req  .eq(0),
                        self.ulpi_data_out .eq(0),
                    ]
                    m.next = 'READ_TURNAROUND'


            # READ_TURNAROUND: wait for the PHY to take control of the ULPI bus.
            with m.State('READ_TURNAROUND'):

                # After one cycle, we should have a data byte ready.
                m.next = 'READ_COMPLETE'


            # READ_COMPLETE: the ULPI read exchange is complete, and the read data is ready.
            with m.State('READ_COMPLETE'):
                m.next = 'IDLE'

                # Latch in the data, and indicate that we have new, valid data.
                m.d.usb += [
                    self.read_data .eq(self.ulpi_data_in),
                    self.done      .eq(1)
                ]

            #
            # Write handling.
            #

            # START_WRITE: wait for the bus to be idle, so we can transmit.
            with m.State('START_WRITE'):

                # Wait for the bus to be idle.
                with m.If(~self.ulpi_dir):
                    m.next = 'SEND_WRITE_ADDRESS'

                    # Once it is, start sending our command.
                    m.d.usb += [
                        self.ulpi_data_out .eq(self.COMMAND_REG_WRITE | self.address),
                        self.ulpi_out_req  .eq(1)
                    ]

            # SEND_WRITE_ADDRESS: Continue sending the write address until the
            # target device accepts it.
            with m.State('SEND_WRITE_ADDRESS'):
                m.d.usb += self.ulpi_out_req.eq(1)

                # If DIR has become asserted, we're being interrupted.
                # We'll have to restart the write after the interruption is over.
                with m.If(self.ulpi_dir):
                    m.next = 'START_WRITE'
                    m.d.usb += self.ulpi_out_req.eq(0)

                # Hold our address until the PHY has accepted the command;
                # and then move to presenting the PHY with the value to be written.
                with m.Elif(self.ulpi_next):
                    m.d.usb += self.ulpi_data_out.eq(self.write_data)
                    m.next = 'HOLD_WRITE'


            # Hold the write data on the bus until the device acknowledges it.
            with m.State('HOLD_WRITE'):
                m.d.usb += self.ulpi_out_req.eq(1)

                # Handle interruption.
                with m.If(self.ulpi_dir):
                    m.next = 'START_WRITE'
                    m.d.usb += self.ulpi_out_req.eq(0)

                # Hold the data present until the device has accepted it.
                # Once it has, pulse STP for a cycle to complete the transaction.
                with m.Elif(self.ulpi_next):
                    m.d.usb += [
                        self.ulpi_data_out.eq(0),
                        self.ulpi_stop.eq(1),
                    ]
                    m.next = 'STOPPING'

            with m.State('STOPPING'):
                m.d.usb += self.ulpi_stop.eq(0)

                # Check again for interruption since DIR may have
                # been asserted during the previous cycle.
                with m.If(self.ulpi_dir):
                    m.next = 'START_WRITE'
                    m.d.usb += self.ulpi_out_req.eq(0)

                with m.Else():
                    m.d.usb += [
                        self.ulpi_out_req.eq(0),
                        self.done.eq(1)
                    ]
                    m.next = 'IDLE'

        return m


class ULPIRxEventDecoder(Elaboratable):
    """ Simple piece of gateware that tracks receive events.

    I/O port:

        I: ulpi_data_in[8] -- The current input state of the ULPI data lines.
        I: ulpi_dir        -- The ULPI bus-direction signal.
        I: ulpi_nxt        -- The ULPI 'next' throttle signal.
        I: register_operation_in_progress
            Signal that should be true iff we're performing a register operation.

        O: last_rx_command -- The full byte value of the last RxCmd.

        O: line_state[2]   -- The states of the two USB lines.
        O: rx_active       -- True when a packet receipt is active.
        O: rx_error        -- True when a packet receive error has occurred.
        O: host_disconnect -- True if the host has just disconnected.
        O: id_digital      -- Digital value of the ID pin.
        O: vbus_valid      -- True iff a valid VBUS voltage is present
        O: session_end     -- True iff a session has just ended.

        # Strobes indicating signal changes.
        O: rx_start        -- True iff an RxEvent has changed the value of RxActive from 0 -> 1.
        O: rx_stop         -- True iff an RxEvent has changed the value of RxActive from 1 -> 0.
    """

    def __init__(self, *, ulpi_bus):

        #
        # I/O port.
        #
        self.ulpi = ulpi_bus
        self.register_operation_in_progress = Signal()

        # Optional: signal that allows access to the last RxCmd byte.
        self.last_rx_command = Signal(8)

        self.line_state      = Signal(2)
        self.rx_active       = Signal()
        self.rx_error        = Signal()
        self.host_disconnect = Signal()
        self.id_digital      = Signal()
        self.vbus_valid      = Signal()
        self.session_valid   = Signal()
        self.session_end     = Signal()

        # RxActive strobes.
        self.rx_start        = Signal()
        self.rx_stop         = Signal()


    def elaborate(self, platform):
        m = Module()

        # An RxCmd is present when three conditions are met:
        # - We're not actively undergoing a register read.
        # - Direction has been high for more than one cycle.
        # - NXT is low.

        # To implement the first condition, we'll first create a delayed
        # version of DIR, and then logically AND it with the current value.
        direction_delayed = Signal()
        m.d.usb += direction_delayed.eq(self.ulpi.dir.i)

        receiving = Signal()
        m.d.comb += receiving.eq(direction_delayed & self.ulpi.dir.i)

        # Default our strobes to 0, unless asserted.
        m.d.usb += [
            self.rx_start  .eq(0),
            self.rx_stop   .eq(0)
        ]

        # Sample the DATA lines whenever these conditions are met.
        with m.If(receiving & ~self.ulpi.nxt.i & ~self.register_operation_in_progress):
            m.d.usb += self.last_rx_command.eq(self.ulpi.data.i)

            # If RxActive has just changed, strobe the start or stop signals,
            rx_active = self.ulpi.data.i[4]
            with m.If(~self.rx_active & rx_active):
                m.d.usb += self.rx_start.eq(1)
            with m.If(self.rx_active & ~rx_active):
                m.d.usb += self.rx_stop.eq(1)


        # Break the most recent RxCmd into its UTMI-equivalent signals.
        # From table 3.8.1.2 in the ULPI spec; rev 1.1/Oct-20-2004.
        m.d.comb += [
            self.line_state      .eq(self.last_rx_command[0:2]),
            self.vbus_valid      .eq(self.last_rx_command[2:4] == 0b11),
            self.session_valid   .eq(self.last_rx_command[2:4] == 0b10),
            self.session_end     .eq(self.last_rx_command[2:4] == 0b00),
            self.rx_active       .eq(self.last_rx_command[4]),
            self.rx_error        .eq(self.last_rx_command[4:6] == 0b11),
            self.host_disconnect .eq(self.last_rx_command[4:6] == 0b10),
            self.id_digital      .eq(self.last_rx_command[6]),
        ]

        return m



class ULPIControlTranslator(Elaboratable):
    """ Gateware that translates ULPI control signals to their UTMI equivalents.

    I/O port:
        I: bus_idle       -- Indicates that the ULPI bus is idle, and thus capable of
                             performing register writes.

        I: xcvr_select[2] -- selects the operating speed of the transciever;
                             00 = HS, 01 = FS, 10 = LS, 11 = LS on FS bus
        I: term_select    -- enables termination for the given operating mode; see spec
        I: op_mode        -- selects the operating mode of the transciever;
                             00 = normal, 01 = non-driving, 10 = disable bit-stuff/NRZI
        I: suspend        -- places the transceiver into suspend mode; active high

        I: id_pullup      -- when set, places a 100kR pull-up on the ID pin
        I: dp_pulldown    -- when set, enables a 15kR pull-down on D+; intended for host mode
        I: dm_pulldown    -- when set, enables a 15kR pull-down on D+; intended for host mode

        I: chrg_vbus      -- when set, connects a resistor from VBUS to GND to discharge VBUS
        I: dischrg_vbus   -- when set, connects a resistor from VBUS to 3V3 to charge VBUS above SessValid

        O: busy           -- true iff the control translator is actively performing an operation
    """

    def __init__(self, *, register_window, own_register_window=False):
        """
        Parmaeters:
            register_window     -- The ULPI register window to work with.
            own_register_window -- True iff we're the owner of this register window.
                                   Typically, we'll use the register window for a broader controller;
                                   but this can be set to True to indicate that we need to consider this
                                   register window our own, and thus a submodule.
        """

        self.register_window     = register_window
        self.own_register_window = own_register_window

        #
        # I/O port
        #
        self.bus_idle      = Signal()
        self.xcvr_select   = Signal(2, init=0b01)
        self.term_select   = Signal()
        self.op_mode       = Signal(2)
        self.suspend       = Signal()

        self.id_pullup     = Signal()
        self.dp_pulldown   = Signal(init=1)
        self.dm_pulldown   = Signal(init=1)

        self.chrg_vbus     = Signal()
        self.dischrg_vbus  = Signal()

        self.busy          = Signal()

        # Extra/non-UTMI properties.
        self.use_external_vbus_indicator = Signal(init=1)

        #
        # Internal variables.
        #
        self._register_signals = {}


    def add_composite_register(self, m, address, value, *, reset_value=0):
        """ Adds a ULPI register that's composed of multiple control signals.

        Params:
            address      -- The register number in the ULPI register space.
            value       -- An 8-bit signal composing the bits that should be placed in
                           the given register.

            reset_value -- If provided, the given value will be assumed as the reset value
                        -- of the given register; allowing us to avoid an initial write.
        """

        current_register_value = Signal(8, init=reset_value, name=f"current_register_value_{address:02x}")

        # Create internal signals that request register updates.
        write_requested = Signal(name=f"write_requested_{address:02x}")
        write_value     = Signal(8, name=f"write_value_{address:02x}")
        write_done      = Signal(name=f"write_done_{address:02x}")

        self._register_signals[address] = {
            'write_requested': write_requested,
            'write_value':     write_value,
            'write_done':      write_done
        }

        # If we've just finished a write, update our current register value.
        with m.If(write_done):
            m.d.usb += current_register_value.eq(write_value),

        # If we have a mismatch between the requested and actual register value,
        # request a write of the new value.
        m.d.comb += write_requested.eq(current_register_value != value)
        with m.If(current_register_value != value):
            m.d.usb += write_value.eq(value)


    def populate_ulpi_registers(self, m):
        """ Creates translator objects that map our control signals to ULPI registers. """

        # Function control.
        function_control = Cat(self.xcvr_select, self.term_select, self.op_mode, Const(0), ~self.suspend, Const(0))
        self.add_composite_register(m, 0x04, function_control, reset_value=0b01000001)

        # OTG control.
        otg_control = Cat(
            self.id_pullup, self.dp_pulldown, self.dm_pulldown, self.dischrg_vbus,
            self.chrg_vbus, Const(0), Const(0), self.use_external_vbus_indicator
        )
        self.add_composite_register(m, 0x0A, otg_control, reset_value=0b00000110)


    def elaborate(self, platform):
        m = Module()

        if self.own_register_window:
            m.submodules.reg_window = self.register_window

        # Add the registers that represent each of our signals.
        self.populate_ulpi_registers(m)

        # Generate logic to handle changes on each of our registers.
        first_element = True
        for address, signals in self._register_signals.items():

            conditional = m.If if first_element else m.Elif
            first_element = False

            # If we're requesting a write on the given register, pass that to our
            # register window.
            with conditional(signals['write_requested']):

                # Keep track of when we'll be okay to start a write:
                # it's when there's a write request, we're not complete.
                # and the bus is idle. We'll use this below.
                request_write = \
                    signals['write_requested'] & \
                    ~self.register_window.done & \
                    self.bus_idle

                m.d.comb += [

                    # Control signals.
                    signals['write_done']              .eq(self.register_window.done),

                    # Register window signals.
                    self.register_window.address       .eq(address),
                    self.register_window.write_data    .eq(signals['write_value']),
                    self.register_window.write_request .eq(request_write),

                    # Status signals
                ]
                m.d.usb += self.busy.eq(request_write | self.register_window.busy)

        # If no register accesses are active, provide default signal values.
        with m.Else():
            m.d.comb += self.register_window.write_request.eq(0)
            m.d.usb  += self.busy.eq(self.register_window.busy)

        # Ensure our register window is never performing a read.
        m.d.comb += self.register_window.read_request.eq(0)

        return m



class ULPITransmitTranslator(Elaboratable):
    """ Accepts UTMI transmit signals, and converts them into ULPI equivalents.

    I/O port:
        I: tx_data[8]      -- The data to be transmitted.
        I: tx_valid        -- Driven high to indicate we're trying to transmit.
        O: tx_ready        -- Driven high when a given byte will be accepted on tx_data on the next clock edge.

        I: op_mode[2]      -- The UTMI operating mode. Used to determine when NOPID commands should be issued;
                              and when to force transmit errors.

        I: bus_idle        -- Should be asserted when the transmitter is able to control the bus.

        O: ulpi_data_out   -- The data to be driven onto the ULPI transmit lines.
        O: ulpi_out_req    -- Asserted when we're trying to drive the ULPI data lines.
        I: ulpi_nxt        -- The NXT signal for the relevant ULPI bus.
        O: ulpi_stp        -- The STP signal for the relevant ULPI bus.

        O: busy            -- True iff this module is using the bus.
    """


    # Prefix for ULPI transmit commands.
    TRANSMIT_COMMAND = 0b01000000

    # UTMI operating mode for "bit stuffing disabled".
    OP_MODE_NO_BIT_STUFFING = 0b10


    def __init__(self):

        #
        # I/O port.
        #

        self.tx_data         = Signal(8)
        self.tx_valid        = Signal()
        self.tx_ready        = Signal()

        self.op_mode         = Signal(2)
        self.bus_idle        = Signal()

        self.ulpi_out_req    = Signal()
        self.ulpi_data_out   = Signal.like(self.tx_data)
        self.ulpi_nxt        = Signal()
        self.ulpi_stp        = Signal()

        self.busy            = Signal()


    def elaborate(self, platform):
        m = Module()
        bit_stuffing_disabled = (self.op_mode == self.OP_MODE_NO_BIT_STUFFING)

        with m.FSM(domain="usb") as fsm:

            # Mark ourselves as busy whenever we're not in idle.
            m.d.comb += self.busy.eq(~fsm.ongoing('IDLE'))

            # IDLE: our transmitter is ready and
            with m.State('IDLE'):
                m.d.comb += self.ulpi_stp.eq(0)

                # Start once a transmit is started, and we can access the bus.
                with m.If(self.tx_valid & self.bus_idle):

                    # If bit-stuffing is disabled, we'll need to prefix our transmission with a NOPID command.
                    # In this case, we'll never accept the first byte (as we're not ready to transmit it, yet),
                    # and thus TxReady will always be 0.
                    with m.If(bit_stuffing_disabled):
                        m.d.usb  += self.ulpi_out_req.eq(1),
                        m.d.comb += [
                            self.ulpi_data_out .eq(self.TRANSMIT_COMMAND),
                            self.tx_ready      .eq(0)
                        ]

                    # Otherwise, this transmission starts with a PID. Extract the PID from the first data byte
                    # and present it as part of the Transmit Command. In this case, the NXT signal is
                    # has the same meaning as the UTMI TxReady signal; and can be passed along directly.
                    with m.Else():
                        m.d.usb  += self.ulpi_out_req.eq(1),
                        m.d.comb += [
                            self.ulpi_data_out .eq(self.TRANSMIT_COMMAND | self.tx_data[0:4]),
                            self.tx_ready      .eq(self.ulpi_nxt)
                        ]


                    # Once the PHY has accepted the command byte, we're ready to move into our main transmit state.
                    with m.If(self.ulpi_nxt):
                        m.next = 'TRANSMIT'


            # TRANSMIT: we're in the body of a transmit; the UTMI and ULPI interface signals
            # are roughly equivalent; we'll just pass them through.
            with m.State('TRANSMIT'):
                m.d.comb += [
                    self.ulpi_data_out .eq(self.tx_data),
                    self.tx_ready      .eq(self.ulpi_nxt),
                    self.ulpi_stp      .eq(0),
                ]

                # Once the transmission has ended, we'll need to issue a ULPI stop.
                with m.If(~self.tx_valid):
                    m.d.usb += self.ulpi_out_req.eq(0),
                    m.next = 'IDLE'

                    # STOP: our packet has just terminated; we'll generate a ULPI stop event for a single cycle.
                    # [ULPI: 3.8.2.2]
                    m.d.comb += self.ulpi_stp.eq(1)

                    # If we've disabled bit stuffing, we'll want to termainate by generating a bit-stuff error.
                    with m.If(bit_stuffing_disabled):

                        # Drive 0xFF as we stop, to generate a bit-stuff error. [ULPI: 3.8.2.3]
                        m.d.comb += self.ulpi_data_out .eq(0xFF)

                    # Otherwise, we'll generate a normal stop.
                    with m.Else():
                        m.d.comb += self.ulpi_data_out .eq(0)


        return m


class UTMITranslator(Elaboratable):
    """ Gateware that translates a ULPI interface into a simpler UTMI one.

    I/O port:

        O: busy          -- signal that's true iff the ULPI interface is being used
                            for a register or transmit command

        # See the UTMI specification for most signals.

        # Data signals:
        I: tx_data[8]  -- data to be transmitted; valid when tx_valid is asserted
        I: tx_valid    -- set to true when data is to be transmitted; indicates the data_in
                          byte is valid; de-asserting this line terminates the transmission
        O: tx_ready    -- indicates the the PHY is ready to accept a new byte of data, and that the
                          transmitter should move on to the next byte after the given cycle

        O: rx_data[8]  -- data received from the PHY; valid when rx_valid is asserted
        O: rx_valid    -- indicates that the data present on rx_data is new and valid data;
                          goes high for a single ULPI clock cycle to indicate new data is ready

        O: rx_active   -- indicates that the PHY is actively receiving data from the host; data is
                          slewed on rx_data by rx_valid
        O: rx_error    -- indicates that an error has occurred in the current transmission

        # Extra signals:
        O: rx_complete -- strobe that goes high for one cycle when a packet rx is complete

        # Signals for diagnostic use:
        O: last_rxcmd    -- The byte content of the last RxCmd.

        I: address       -- The ULPI register address to work with.
        O: read_data[8]  -- The contents of the most recently read ULPI command.
        I: write_data[8] -- The data to be written on the next write request.
        I: manual_read   -- Strobe that triggers a diagnostic read.
        I: manual_write  -- Strobe that triggers a diagnostic write.

    """

    _CYCLES_1_MILLISECONDS = 60000

    # UTMI status signals translated from the ULPI bus.
    RXEVENT_STATUS_SIGNALS = [
        ('line_state', 2), ('vbus_valid', 1), ('session_valid', 1), ('session_end', 1),
        ('rx_error',   1), ('host_disconnect', 1), ('id_digital', 1)
    ]

    # Control signals that we control through our control translator.
    CONTROL_SIGNALS = [
        ('xcvr_select',  2), ('term_select', 1), ('op_mode',     2), ('suspend',   1),
        ('id_pullup',    1), ('dm_pulldown', 1), ('dp_pulldown', 1), ('chrg_vbus', 1),
        ('dischrg_vbus', 1), ('use_external_vbus_indicator', 1)
    ]


    def __dir__(self):
        """ Extend our properties list of contain all of the above fields, for proper autocomplete. """

        properties = list(super().__dir__())

        properties.extend(name for name, _ in self.RXEVENT_STATUS_SIGNALS)
        properties.extend(name for name, _ in self.CONTROL_SIGNALS)

        return properties


    def __init__(self, *, ulpi, use_platform_registers=True, handle_clocking=True):
        """ Params:

            ulpi                   -- The ULPI bus to communicate with.
            use_platform_registers -- If True (or not provided), any extra registers writes provided in
                                      the platform definition will be applied automatically.
            handle_clocking        -- True iff we should attempt to automatically handle ULPI clocking. If
                                      the `clk` ULPI signal is an input, it will be used to provide the 'usb'
                                      domain clock. If the ULPI signal is an output, it will driven with our
                                      'usb' domain clock. If False, it will be the user's responsibility to
                                      handle clocking.

            Note that it's recommended that multi-PHY systems either use a single clock for all PHYs
            (assuming the PHYs support clock input), or that individual clock domains be created for each
            PHY using a DomainRenamer.
        """

        self.use_platform_registers = use_platform_registers
        self.handle_clocking        = handle_clocking

        #
        # I/O port
        #
        self.ulpi            = ulpi
        self.busy            = Signal()

        # Data signals.
        self.rx_data         = Signal(8)
        self.rx_valid        = Signal()

        self.tx_data         = Signal(8)
        self.tx_valid        = Signal()
        self.tx_ready        = Signal()

        # Status signals.
        self.rx_active       = Signal()

        # RxEvent-based flags.
        for signal_name, size in self.RXEVENT_STATUS_SIGNALS:
            self.__dict__[signal_name] = Signal(size, name=signal_name)

        # Control signals.
        for signal_name, size in self.CONTROL_SIGNALS:
            self.__dict__[signal_name] = Signal(size, name=signal_name)

        # Diagnostic I/O.
        self.last_rx_command = Signal(8)

        #
        # Internal
        #

        #  Create a list of extra registers to be set.
        self._extra_registers = {}


    def add_extra_register(self, write_address, write_value, *, default_value=None):
        """ Adds logic to configure an extra ULPI register. Useful for configuring vendor registers.

        Params:
            write_address -- The write address of the target ULPI register.
            write_value   -- The value to be written. If a Signal is provided; the given register will be
                             set post-reset, if necessary; and then dynamically updated each time the signal changes.
                             If an integer constant is provided, this value will be written once upon startup.
            default_value -- The default value the register is expected to have post-reset; used to determine
                             if the value needs to be updated post-reset. If a Signal is provided for write_value,
                             this must be provided; if an integer is provided for write_value, this is optional.
        """

        # Ensure we have a default_value if we have a Signal(); as this will determine
        # whether we need to update the register post-reset.
        if (default_value is None) and isinstance(write_value, Signal):
            raise ValueError("if write_value is a signal, default_value must be provided")

        # Otherwise, we'll pick a value that ensures the write always occurs.
        elif default_value is None:
            default_value = 0xff ^ write_value

        self._extra_registers[write_address] = {'value': write_value, 'default': default_value}


    def elaborate(self, platform):
        m = Module()

        # Create the component parts of our ULPI interfacing hardware.
        m.submodules.register_window     = register_window     = ULPIRegisterWindow()
        m.submodules.control_translator  = control_translator  = ULPIControlTranslator(register_window=register_window)
        m.submodules.rxevent_decoder     = rxevent_decoder     = ULPIRxEventDecoder(ulpi_bus=self.ulpi)
        m.submodules.transmit_translator = transmit_translator = ULPITransmitTranslator()

        # If we're choosing to honor any registers defined in the platform file, apply those
        # before continuing with elaboration.
        if self.use_platform_registers and hasattr(platform, 'ulpi_extra_registers'):
            for address, value in platform.ulpi_extra_registers.items():
                self.add_extra_register(address, value)

        # Some platforms may need to have a raw clock domain for their I/O; e.g. if they need
        # to do some simple processing for their internal clock domain.
        if self.handle_clocking and hasattr(platform, 'ulpi_raw_clock_domain'):
            raw_clock_domain = platform.ulpi_raw_clock_domain
        else:
            raw_clock_domain = 'usb'


        # Add any extra registers provided by the user to our control translator.
        for address, values in self._extra_registers.items():
            control_translator.add_composite_register(m, address, values['value'], reset_value=values['default'])

        # Keep track of when any of our components are busy
        any_busy = \
            register_window.busy     | \
            transmit_translator.busy | \
            control_translator.busy  | \
            self.ulpi.dir.i


        # If we're handling ULPI clocking, do so.
        if self.handle_clocking:

            # We can't currently handle bidirectional clock lines, as we don't know if they
            # should be used in input or output modes.
            if hasattr(self.ulpi.clk, 'oe'):
                raise TypeError("ULPI records with bidirectional clock lines require manual handling.")

            # Just Input (TM) and Just Output (TM) clocks are simpler: we know how to drive them.
            elif hasattr(self.ulpi.clk, 'o'):
                m.d.comb += self.ulpi.clk.o.eq(ClockSignal(raw_clock_domain))
            elif hasattr(self.ulpi.clk, 'i'):
                m.d.comb += ClockSignal(raw_clock_domain).eq(self.ulpi.clk.i)

            # Clocks that don't seem to be I/O pins aren't what we're expecting; fail out.
            else:
                raise TypeError(f"ULPI `clk` was an unexpected type {type(self.ulpi.clk)}." \
                    " You may need to handle clocking manually.")


        # Hook up our reset signal iff our ULPI bus has one.
        phy_ready = Signal()
        if hasattr(self.ulpi, 'rst'):
            m.d.comb += self.ulpi.rst.o.eq(ResetSignal(raw_clock_domain)),

            # After reset, DIR may not be driven high immediately.
            # Before using the bus, wait for the minimum Tstart time according to [USB334x: Table 4.3].
            startup_counter = Signal(range(self._CYCLES_1_MILLISECONDS + 1))
            m.d.usb += startup_counter.eq(startup_counter + 1)
            with m.If(startup_counter == self._CYCLES_1_MILLISECONDS):
                m.d.usb += phy_ready.eq(1)
        else:
            m.d.usb += phy_ready.eq(1)


        # Connect our ULPI control signals to each of our subcomponents.
        m.d.comb += [

            # Drive the bus whenever the target PHY isn't.
            self.ulpi.data.oe            .eq(~self.ulpi.dir.i),

            # Generate our busy signal.
            self.busy                    .eq(any_busy),

            # Connect our data inputs to the event decoder.
            # Note that the event decoder is purely passive.
            rxevent_decoder.register_operation_in_progress.eq(register_window.busy),
            self.last_rx_command          .eq(rxevent_decoder.last_rx_command),

            # Connect our inputs to our transmit translator.
            transmit_translator.ulpi_nxt  .eq(self.ulpi.nxt.i),
            transmit_translator.op_mode   .eq(self.op_mode),
            transmit_translator.bus_idle  .eq(~control_translator.busy & ~self.ulpi.dir.i & phy_ready),
            transmit_translator.tx_data   .eq(self.tx_data),
            transmit_translator.tx_valid  .eq(self.tx_valid),
            self.tx_ready                 .eq(transmit_translator.tx_ready),

            # Connect our inputs to our control translator / register window.
            control_translator.bus_idle   .eq(~transmit_translator.busy & phy_ready),
            register_window.ulpi_data_in  .eq(self.ulpi.data.i),
            register_window.ulpi_dir      .eq(self.ulpi.dir.i),
            register_window.ulpi_next     .eq(self.ulpi.nxt.i),
        ]

        # Control our the source of our ULPI data output.
        # If a transmit request is active, prioritize that over
        # any register reads/writes.
        with m.If(transmit_translator.ulpi_out_req):
            m.d.comb += [
                self.ulpi.data.o  .eq(transmit_translator.ulpi_data_out),
                self.ulpi.stp.o   .eq(transmit_translator.ulpi_stp)
            ]
        # Otherwise, yield control to the register handler.
        # This is a slight optimization: since it properly generates NOPs
        # while not in use, we can let it handle idle, as well, saving a mux.
        with m.Else():
            m.d.comb += [
                self.ulpi.data.o  .eq(register_window.ulpi_data_out),
                self.ulpi.stp.o   .eq(register_window.ulpi_stop)
            ]


        # Connect our RxEvent status signals from our RxEvent decoder.
        for signal_name, _ in self.RXEVENT_STATUS_SIGNALS:
            signal = getattr(rxevent_decoder, signal_name)
            m.d.comb += self.__dict__[signal_name].eq(signal)

        # Connect our control signals through the control translator.
        for signal_name, _ in self.CONTROL_SIGNALS:
            signal = getattr(control_translator, signal_name)
            m.d.comb += signal.eq(self.__dict__[signal_name])


        # RxActive handler:
        # A transmission starts when DIR goes high with NXT, or when an RxEvent indicates
        # a switch from RxActive = 0 to RxActive = 1. A transmission stops when DIR drops low,
        # or when the RxEvent RxActive bit drops from 1 to 0, or an error occurs.A
        past_dir        = Signal.like(self.ulpi.dir.i)
        m.d.usb        += past_dir.eq(self.ulpi.dir.i)
        dir_rising_edge = ~past_dir & self.ulpi.dir.i
        dir_based_start = dir_rising_edge & self.ulpi.nxt.i


        with m.If(~self.ulpi.dir.i | rxevent_decoder.rx_stop):
            # TODO: this should probably also trigger if RxError
            m.d.usb += self.rx_active.eq(0)
        with m.Elif(dir_based_start | rxevent_decoder.rx_start):
            m.d.usb += self.rx_active.eq(1)


        # Data-out: we'll connect this almost direct through from our ULPI
        # interface, as it's essentially the same as in the UTMI spec. We'll
        # add a one cycle processing delay so it matches the rest of our signals.

        # RxValid: equivalent to NXT whenever a Rx is active.
        m.d.usb += [
            self.rx_data   .eq(self.ulpi.data.i),
            self.rx_valid  .eq(self.ulpi.nxt.i & self.rx_active)
        ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test import LunaGatewareTestCase, sync_test_case

from luna.gateware.memory import TransactionalizedFIFO

class TransactionalizedFIFOTest(LunaGatewareTestCase):
    FRAGMENT_UNDER_TEST = TransactionalizedFIFO
    FRAGMENT_ARGUMENTS = {'width': 8, 'depth': 16}

    def initialize_signals(self):
        yield self.dut.write_en.eq(0)

    @sync_test_case
    def test_simple_fill(self):
        dut = self.dut

        # Our FIFO should start off empty; and with a full depth of free space.
        self.assertEqual((yield dut.empty),           1)
        self.assertEqual((yield dut.full),            0)
        self.assertEqual((yield dut.space_available), 16)

        # If we add a byte to the queue...
        yield dut.write_data.eq(0xAA)
        yield from self.pulse(dut.write_en)

        # ... we should have less space available ...
        self.assertEqual((yield dut.space_available), 15)

        # ... but we still should be "empty", as we won't have data to read until we commit.
        self.assertEqual((yield dut.empty), 1)

        # Once we _commit_ our write, we should suddenly have data to read.
        yield from self.pulse(dut.write_commit)
        self.assertEqual((yield dut.empty), 0)

        # If we read a byte, we should see the FIFO become empty...
        yield from self.pulse(dut.read_en)
        self.assertEqual((yield dut.empty), 1)

        # ... but we shouldn't see more space become available until we commit the read.
        self.assertEqual((yield dut.space_available), 15)
        yield from self.pulse(dut.read_commit)
        self.assertEqual((yield dut.space_available), 16)

        # If we write 16 more bytes of data...
        yield dut.write_en.eq(1)
        for i in range(16):
            yield dut.write_data.eq(i)
            yield
        yield dut.write_en.eq(0)

        # ... our buffer should be full, but also empty.
        # This paradox exists as we've filled our buffer with uncomitted data.
        yield
        self.assertEqual((yield dut.full),  1)
        self.assertEqual((yield dut.empty), 1)

        # Once we _commit_ our data, it should suddenly stop being empty.
        yield from self.pulse(dut.write_commit)
        self.assertEqual((yield dut.empty), 0)

        # Reading a byte _without committing_ shouldn't change anything about empty/full/space-available...
        yield from self.pulse(dut.read_en)
        self.assertEqual((yield dut.empty), 0)
        self.assertEqual((yield dut.full),  1)
        self.assertEqual((yield dut.space_available),  0)

        # ... but committing should increase our space available by one, and make our buffer no longer full.
        yield from self.pulse(dut.read_commit)
        self.assertEqual((yield dut.empty), 0)
        self.assertEqual((yield dut.full),  0)
        self.assertEqual((yield dut.space_available),  1)

        # Reading/committing another byte should increment our space available.
        yield from self.pulse(dut.read_en)
        yield from self.pulse(dut.read_commit)
        self.assertEqual((yield dut.space_available),  2)

        # Writing data into the buffer should then fill it back up again...
        yield dut.write_en.eq(1)
        for i in range(2):
            yield dut.write_data.eq(i)
            yield
        yield dut.write_en.eq(0)

        # ... meaning it will again be full, and have no space remaining.
        yield
        self.assertEqual((yield dut.full),             1)
        self.assertEqual((yield dut.space_available),  0)

        # If we _discard_ this data, we should go back to having two bytes available.
        yield from self.pulse(dut.write_discard)
        self.assertEqual((yield dut.full),             0)
        self.assertEqual((yield dut.space_available),  2)

        # If we read the data that's remaining in the fifo...
        yield dut.read_en.eq(1)
        for i in range(2, 16):
            yield
            self.assertEqual((yield dut.read_data), i)
        yield dut.read_en.eq(0)

        # ... our buffer should again be empty.
        yield
        self.assertEqual((yield dut.empty),            1)
        self.assertEqual((yield dut.space_available),  2)

        # If we _discard_ our current read, we should then see our buffer no longer empty...
        yield from self.pulse(dut.read_discard)
        self.assertEqual((yield dut.empty),            0)

        # and we should be able to read the same data again.
        yield dut.read_en.eq(1)
        for i in range(2, 16):
            yield
            self.assertEqual((yield dut.read_data), i)
        yield dut.read_en.eq(0)

        # On committing this, we should see a buffer that is no longer full, and is really empty.
        yield from self.pulse(dut.read_commit)
        self.assertEqual((yield dut.empty),            1)
        self.assertEqual((yield dut.full),             0)
        self.assertEqual((yield dut.space_available),  16)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Import shortcuts for our most commonly used functionality. """

# Create shorthands for the most common parts of the library's usb3 gateware.
from .gateware.usb.usb3.device              import USBSuperSpeedDevice
from .gateware.usb.usb3.application.request import SuperSpeedRequestHandlerInterface, SuperSpeedRequestHandler
from .gateware.usb.usb3.endpoints.stream    import SuperSpeedStreamInEndpoint

__all__ = ['USBSuperSpeedDevice', 'SuperSpeedRequestHandler']



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Low-level USB transciever gateware -- control transfer components. """

from amaranth              import Signal, Module, Elaboratable
from usb_protocol.emitters import DeviceDescriptorCollection
from usb_protocol.types    import USBRequestType

from .packet               import DataCRCInterface, USBDataPacketCRC, USBInterpacketTimer
from .packet               import USBTokenDetector, TokenDetectorInterface
from .packet               import InterpacketTimerInterface, HandshakeExchangeInterface
from .endpoint             import EndpointInterface
from .request              import USBSetupDecoder, USBRequestHandlerMultiplexer
from ..request.standard    import StandardRequestHandler
from ..stream              import USBInStreamInterface, USBOutStreamInterface


class USBControlEndpoint(Elaboratable):
    """ Gateware that manages control request data progression.

    This class is used by creating one or more *request handler* modules; which define how requests
    are handled. These handlers can be bound using :attr:`add_request_handler`.

    For convenience, this module can also automatically be populated with a ``StandardRequestHandler``
    via the :attr:`add_standard_request_handlers`.

    Attributes
    ----------
    interface: EndpointInterface
        The interface from this endpoint to the core device hardware.

    Parameters
    ----------
        utmi: UTMI bus, or equivalent translator
            The UTMI bus we'll monitor for data. We'll consider this read-only.
        endpoint_number: int, optional
            The endpoint number for this control interface; defaults to (and almost always should
            be) zero.
        standalone: bool
            Debug parameter. If true, this module will operate without external components;
            i.e. without an internal data-CRC generator, or tokenizer. In this case, tokenizer
            and timer should be set to None; and will be ignored.
    """

    def __init__(self, *, utmi, endpoint_number=0, standalone=False, max_packet_size=64):
        self.utmi             = utmi
        self._standalone      = standalone
        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size

        #
        # I/O Port
        #
        self.interface = EndpointInterface()

        #
        # Internals.
        #

        # List of the modules that will handle control requests.
        self._request_handlers = []


    def add_request_handler(self, request_handler):
        """ Adds a ControlRequestHandler module to this control endpoint.

        No arbitration is performed between request handlers; so it's important
        that request handlers not overlap in the requests they handle.
        """
        self._request_handlers.append(request_handler)


    def add_standard_request_handlers(self, descriptors: DeviceDescriptorCollection, **kwargs):
        """ Adds a handlers for the standard USB requests.

        This will handle all Standard-type requests; so any additional request handlers
        must not handle Standard requests.

        Parameters will be passed on to StandardRequestHandler.
        """
        handler = StandardRequestHandler(descriptors, max_packet_size=self._max_packet_size, **kwargs)
        self._request_handlers.append(handler)


    def _handle_setup_reset(self, m):
        """ Adds a FSM condition that moves back to the SETUP phase if we ever receive a setup token.

        Should only be used within our core FSM.
        """
        tokenizer = self.interface.tokenizer

        # If we receive a SETUP token, always move back to the SETUP stage.
        with m.If(tokenizer.new_token & tokenizer.is_setup):
            m.next = 'SETUP'


    def elaborate(self, platform):
        m = Module()
        interface = self.interface

        #
        # Test scaffolding.
        #

        if self._standalone:

            # Create our timer...
            m.submodules.timer = timer = USBInterpacketTimer()
            timer.add_interface(interface.timer)

            # ... our CRC generator ...
            m.submodules.crc = crc = USBDataPacketCRC()
            crc.add_interface(interface.data_crc)
            m.d.comb += [
                crc.rx_data    .eq(self.utmi.rx_data),
                crc.rx_valid   .eq(self.utmi.rx_valid),
                crc.tx_valid   .eq(0)
            ]

            # ... and our tokenizer.
            m.submodules.token_detector = tokenizer = USBTokenDetector(utmi=self.utmi)
            m.d.comb += tokenizer.interface.connect(interface.tokenizer)


        #
        # Submodules
        #

        # Create our SETUP packet decoder.
        m.submodules.setup_decoder = setup_decoder = USBSetupDecoder(utmi=self.utmi)
        m.d.comb += [
            interface.data_crc   .connect(setup_decoder.data_crc),
            interface.tokenizer  .connect(setup_decoder.tokenizer),
            setup_decoder.speed  .eq(interface.speed),

            # And attach our timer interface to both our local users and
            # to our setup decoder.
            interface.timer      .attach(setup_decoder.timer)

        ]


        #
        # Request handler logic.
        #

        # Multiplex the output of each of our request handlers.
        m.submodules.request_mux = request_mux = USBRequestHandlerMultiplexer()
        request_handler = request_mux.shared

        # Add each of our handlers to the endpoint; and add it to our mux.
        for handler in self._request_handlers:

            # Create a display name for the handler...
            name = handler.__class__.__name__
            if hasattr(m.submodules, name):
                name = f"{name}_{id(handler)}"

            # ... and add it.
            m.submodules[name] = handler
            request_mux.add_interface(handler.interface)


        # ... and hook it up.
        m.d.comb += [
            setup_decoder.packet                   .connect(request_handler.setup),
            interface.tokenizer                    .connect(request_handler.tokenizer),

            request_handler.tx                     .attach(interface.tx),
            interface.handshakes_out.ack           .eq(setup_decoder.ack | request_handler.handshakes_out.ack),
            interface.handshakes_out.nak           .eq(request_handler.handshakes_out.nak),
            interface.handshakes_out.stall         .eq(request_handler.handshakes_out.stall),
            interface.handshakes_in                .connect(request_handler.handshakes_in),

            interface.address_changed              .eq(request_handler.address_changed),
            interface.new_address                  .eq(request_handler.new_address),

            interface.clear_endpoint_halt_out      .eq(request_handler.clear_endpoint_halt),

            request_handler.active_config          .eq(interface.active_config),
            interface.config_changed               .eq(request_handler.config_changed),
            interface.new_config                   .eq(request_handler.new_config),

            # Fix our data PIDs to DATA1, for now, as we don't support multi-packet responses, yet.
            # Per [USB2.0: 8.5.3], the first packet of the DATA or STATUS phase always carries a DATA1 PID.
            interface.tx_pid_toggle                .eq(request_handler.tx_data_pid)
        ]


        #
        # Core control request handler.
        # Behavior dictated by [USB2, 8.5.3].
        #
        endpoint_targeted = (self.interface.tokenizer.endpoint == self._endpoint_number)
        with m.FSM(domain="usb"):

            # SETUP -- The "SETUP" phase of a control request. We'll wait here
            # until the SetupDetector detects a valid setup packet for us.
            with m.State('SETUP'):

                # We won't do anything until we receive a SETUP token.
                with m.If(setup_decoder.packet.received & endpoint_targeted):

                    # If our SETUP packet indicates we'll have a data stage (wLength > 0)
                    # move to the DATA stage. Otherwise, move directly to the status stage [8.5.3].
                    with m.If(setup_decoder.packet.length):

                        # If this is an device -> host request, expect an IN packet.
                        with m.If(setup_decoder.packet.is_in_request):
                            m.next = 'DATA_IN'

                        # Otherwise, expect an OUT one.
                        with m.Else():
                            m.next = 'DATA_OUT'

                    with m.Else():
                        # If we don't have a data phase, our status phase is always an IN [USB2.0: 8.5.3]
                        m.next = 'STATUS_IN'


            with m.State('DATA_IN'):
                self._handle_setup_reset(m)

                # Wait until we have an IN token, and are allowed to respond to it.
                allowed_to_respond = interface.tokenizer.ready_for_response & endpoint_targeted
                with m.If(allowed_to_respond & interface.tokenizer.is_in):

                    # Notify the request handler to prepare a response.
                    m.d.comb += request_handler.data_requested.eq(1)

                # Once we get an OUT token, we should move on to the STATUS stage. [USB2, 8.5.3]
                with m.If(endpoint_targeted & interface.tokenizer.new_token & (interface.tokenizer.is_out | interface.tokenizer.is_ping)):
                    m.next = 'STATUS_OUT'


            with m.State('DATA_OUT'):
                self._handle_setup_reset(m)

                # Pass through our Rx related signals iff we're in the DATA_OUT stage,
                # and the most recent token pointed to our endpoint. This ensures the
                # request handler only ever sees data events related to it; this simplifies
                # the request handler logic significantly.
                with m.If(endpoint_targeted & interface.tokenizer.is_out):
                    m.d.comb += [
                        interface.rx                           .connect(request_handler.rx),
                        request_handler.rx_ready_for_response  .eq(interface.rx_ready_for_response),
                        request_handler.rx_invalid             .eq(interface.rx_invalid),
                    ]

                # Once we get an IN token, we should move on to the STATUS stage. [USB2, 8.5.3]
                with m.If(endpoint_targeted & interface.tokenizer.new_token & interface.tokenizer.is_in):
                    m.next = 'STATUS_IN'

                # Respond to PING token [USB2.0: 8.5.1]
                with m.If(endpoint_targeted & interface.tokenizer.ready_for_response & interface.tokenizer.is_ping):
                    m.d.comb += interface.handshakes_out.ack.eq(1)


            # STATUS_IN -- We're currently in the status stage, and we're expecting an IN token.
            # We'll wait for that token.
            with m.State('STATUS_IN'):
                self._handle_setup_reset(m)

                # If we respond to a status-phase IN token, we'll always use a DATA1 PID [USB2.0: 8.5.3]

                # When we get an IN token, the host is looking for a status-stage ZLP.
                # Notify the target handler.
                allowed_to_respond = interface.tokenizer.ready_for_response & endpoint_targeted
                with m.If(allowed_to_respond & interface.tokenizer.is_in):
                    m.d.comb += request_handler.status_requested.eq(1)


            # STATUS_OUT -- We're currently in the status stage, and we're expecting the DATA packet for
            # an OUT request.
            with m.State('STATUS_OUT'):
                self._handle_setup_reset(m)

                # Once we've received a new DATA packet, we're ready to handle a status request.
                allowed_to_respond = interface.rx_ready_for_response & endpoint_targeted
                with m.If(allowed_to_respond & interface.tokenizer.is_out):
                    m.d.comb += request_handler.status_requested.eq(1)

                # Respond to PING token [USB2.0: 8.5.1]
                with m.If(endpoint_targeted & interface.tokenizer.ready_for_response & interface.tokenizer.is_ping):
                    m.d.comb += interface.handshakes_out.ack.eq(1)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Gateware for working with abstract endpoints. """

import functools
import operator

from amaranth         import Signal, Elaboratable, Module, Cat

from .packet          import DataCRCInterface, InterpacketTimerInterface, TokenDetectorInterface
from .packet          import HandshakeExchangeInterface
from .request         import ClearEndpointHaltInterface
from ..stream         import USBInStreamInterface, USBOutStreamInterface
from ...utils.bus     import OneHotMultiplexer


class EndpointInterface:
    """ Interface that connects a USB endpoint module to a USB device.

    Many non-control endpoints won't need to use the latter half of this structure;
    it will be automatically removed by the relevant synthesis tool.

    Attributes
    ----------
    tokenizer: TokenDetectorInterface, to detector
        Interface to our TokenDetector; notifies us of USB tokens.

    rx: USBOutStreamInterface, input stream to endpoint
        Receive interface for this endpoint.
    rx_complete: Signal(), input to endpoint
        Strobe that indicates that the concluding rx-stream was valid (CRC check passed).
    rx_ready_for_response: Signal(), input to endpoint
        Strobe that indicates that we're ready to respond to a complete transmission.
        Indicates that an interpacket delay has passed after an `rx_complete` strobe.
    rx_invalid: Signal(), input to endpoint
        Strobe that indicates that the concluding rx-stream was invalid (CRC check failed).
    rx_pid_toggle: Signal(), input to endpoint
        Value for the data PID toggle; 0 indicates we're receiving a DATA0; 1 indicates Data1.

    tx: USBInStreamInterface, output stream from endpoint
        Transmit interface for this endpoint.
    tx_pid_toggle: Signal(2), output from endpoint
        Value for the data PID toggle; 0 indicates we'll send DATA0; 1 indicates DATA1.
        2 indicates we'll send DATA2, while 3 indicates we'll send DATAM.

    handshakes_in: HandshakeExchangeInterface, input to endpoint
        Carries handshakes detected from the host.
    handshakes_out: HandshakeExchangeInterface, output from endpoint
        Carries handshakes generate by this endpoint.

    speed: Signal(2), input to endpoint
        The device's current operating speed. Should be a USBSpeed enumeration value --
        0 for high, 1 for full, 2 for low.

    active_address: Signal(7), input to endpoint
        Contains the device's current address.
    address_changed: Signal(), output from endpoint.
        Strobe; pulses high when the device's address should be changed.
    new_address: Signal(7), output from endpoint
        When :attr:`address_changed` is high, this field contains the address that should be adopted.

    active_config: Signal(8), input to endpoint
        The configuration number of the active configuration.
    config_changed: Signal(), output from endpoint
        Strobe; pulses high when the device's configuration should be changed.
    new_config: Signal(8)
        When `config_changed` is high, this field contains the configuration that should be applied.

    timer: InterpacketTimerInterface
        Interface to our interpacket timer.
    data_crc: DataCRCInterface
        Control connection for our data-CRC unit.
    """

    def __init__(self):
        self.data_crc              = DataCRCInterface()
        self.tokenizer             = TokenDetectorInterface()
        self.timer                 = InterpacketTimerInterface()

        self.speed                 = Signal(2)

        self.active_address        = Signal(7)
        self.address_changed       = Signal()
        self.new_address           = Signal(7)

        self.active_config         = Signal(8)
        self.config_changed        = Signal()
        self.new_config            = Signal(8)

        self.clear_endpoint_halt_out = Signal(ClearEndpointHaltInterface)
        self.clear_endpoint_halt_in  = Signal(ClearEndpointHaltInterface)

        self.rx                    = USBOutStreamInterface()
        self.rx_complete           = Signal()
        self.rx_ready_for_response = Signal()
        self.rx_invalid            = Signal()
        self.rx_pid_toggle         = Signal(2)

        self.tx                    = USBInStreamInterface()
        self.tx_pid_toggle         = Signal(2)

        self.handshakes_in         = HandshakeExchangeInterface(is_detector=True)
        self.handshakes_out        = HandshakeExchangeInterface(is_detector=False)
        self.issue_stall           = Signal()


class USBEndpointMultiplexer(Elaboratable):
    """ Multiplexes access to the resources shared between multiple endpoint interfaces.

    Interfaces are added using :attr:`add_interface`.

    Attributes
    ----------

    shared: EndpointInterface
        The post-multiplexer endpoint interface.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.shared = EndpointInterface()

        #
        # Internals
        #
        self._interfaces = []


    def add_interface(self, interface: EndpointInterface):
        """ Adds a EndpointInterface to the multiplexer.

        Arbitration is not performed; it's expected only one endpoint will be
        driving the transmit lines at a time.
        """
        self._interfaces.append(interface)


    def _multiplex_signals(self, m, *, when, multiplex, sub_bus=None):
        """ Helper that creates a simple priority-encoder multiplexer.

        Parmeters
        ---------
        when: str
            The name of the interface signal that indicates that the `multiplex` signals should be
            selected for output. If this signals should be multiplexed, it should be included in `multiplex`.
        multiplex: iterable(str)
            The names of the interface signals to be multiplexed.
        """

        def get_signal(interface, name):
            """ Fetches an interface signal by name / sub_bus. """

            if sub_bus:
                bus = getattr(interface, sub_bus)
                return getattr(bus, name)
            else:
                return  getattr(interface, name)


        # We're building an if-elif tree; so we should start with an If entry.
        conditional = m.If

        for interface in self._interfaces:
            condition = get_signal(interface, when)

            with conditional(condition):

                # Connect up each of our signals.
                for signal_name in multiplex:

                    # Get the actual signals for our input and output...
                    driving_signal = get_signal(interface,   signal_name)
                    target_signal  = get_signal(self.shared, signal_name)

                    # ... and connect them.
                    m.d.comb += target_signal   .eq(driving_signal)

            # After the first element, all other entries should be created with Elif.
            conditional = m.Elif


    def or_join_interface_signals(self, m, signal_for_interface):
        """ Joins together a set of signals on each interface by OR'ing the signals together. """

        # Find the value of all of our pre-mux signals OR'd together...
        all_signals = (signal_for_interface(i) for i in self._interfaces)
        or_value = functools.reduce(operator.__or__, all_signals, 0)

        # ... and tie it to our post-mux signal.
        m.d.comb += signal_for_interface(self.shared).eq(or_value)


    def elaborate(self, platform):
        m = Module()
        shared = self.shared

        #
        # Pass through signals being routed -to- our pre-mux interfaces.
        #
        for interface in self._interfaces:
            m.d.comb += [

                # CRC and timer shared signals interface.
                interface.data_crc.crc           .eq(shared.data_crc.crc),
                interface.timer.tx_allowed       .eq(shared.timer.tx_allowed),
                interface.timer.tx_timeout       .eq(shared.timer.tx_timeout),
                interface.timer.rx_timeout       .eq(shared.timer.rx_timeout),

                # Detectors.
                shared.handshakes_in             .connect(interface.handshakes_in),
                shared.tokenizer                 .connect(interface.tokenizer),

                interface.clear_endpoint_halt_in .eq(shared.clear_endpoint_halt_out),

                # Rx interface.
                shared.rx                        .connect(interface.rx),
                interface.rx_complete            .eq(shared.rx_complete),
                interface.rx_ready_for_response  .eq(shared.rx_ready_for_response),
                interface.rx_invalid             .eq(shared.rx_invalid),
                interface.rx_pid_toggle          .eq(shared.rx_pid_toggle),

                # State signals.
                interface.speed                  .eq(shared.speed),
                interface.active_config          .eq(shared.active_config),
                interface.active_address         .eq(shared.active_address)
            ]

        #
        # Multiplex the signals being routed -from- our pre-mux interface.
        #
        self._multiplex_signals(m,
            when='address_changed',
            multiplex=['address_changed', 'new_address']
        )
        self._multiplex_signals(m,
            when='config_changed',
            multiplex=['config_changed', 'new_config']
        )

        # Connect up our transmit interface.
        m.submodules.tx_mux = tx_mux = OneHotMultiplexer(
            interface_type=USBInStreamInterface,
            mux_signals=('payload',),
            or_signals=('valid', 'first', 'last'),
            pass_signals=('ready',)
        )
        tx_mux.add_interfaces(i.tx for i in self._interfaces)
        m.d.comb += self.shared.tx.stream_eq(tx_mux.output)

        # OR together all of our handshake-generation requests...
        self.or_join_interface_signals(m, lambda interface : interface.handshakes_out.ack)
        self.or_join_interface_signals(m, lambda interface : interface.handshakes_out.nak)
        self.or_join_interface_signals(m, lambda interface : interface.handshakes_out.stall)

        # ... our CRC start signals...
        self.or_join_interface_signals(m, lambda interface : interface.data_crc.start)

        # ... and our timer start signals.
        self.or_join_interface_signals(m, lambda interface : interface.timer.start)

        self.or_join_interface_signals(m, lambda interface : interface.clear_endpoint_halt_out.enable)
        self.or_join_interface_signals(m, lambda interface : interface.clear_endpoint_halt_out.direction)
        self.or_join_interface_signals(m, lambda interface : interface.clear_endpoint_halt_out.number)

        # Finally, connect up our transmit PID select.
        conditional = m.If

        # We'll connect our PID toggle to whichever interface has a valid transmission going.
        past_valid  = Signal(len(self._interfaces))
        m.d.usb    += past_valid.eq(Cat(interface.tx.valid for interface in self._interfaces))
        for i, interface in enumerate(self._interfaces):
            with conditional(interface.tx.valid | past_valid[i]):
                m.d.comb += shared.tx_pid_toggle.eq(interface.tx_pid_toggle)

            conditional = m.Elif

        return m



#
# This file is part of LUNA.
#
""" Pure-gateware, UTMI-compatible Full Speed PHY."""

import logging

from amaranth         import Signal, Module, Cat, Elaboratable, ClockSignal

from .receiver      import RxPipeline
from .transmitter   import TxPipeline

class GatewarePHY(Elaboratable):
    """ Gateware that implements a UTMI-compatible transciever using raw FPGA I/O.

    Clock Domains
    -------------

    usb:
        Our 12 MHz USB clock domain which will match each of the signals below.
        Should be phase-related to the 48MHz clock (e.g. divided down from it) to avoid
        the need for explicit synchronization on clock domain crossings.
    usb_io:
        The core 48MHz clock domain in which USB clock recovery and sampling is performed.
        Must be phase related to our ``usb`` clock domain.

    Attributes
    ----------

    tx_data: Signal(8), input
        data to be transmitted; valid when tx_valid is asserted
    tx_valid: Signal(), input
        asserted when data is to be transmitted; indicates the data_in byte is valid;
        de-asserting this line terminates the transmission
    tx_ready: Signal(), output
        indicates the the PHY is ready to accept a new byte of data, and that the transmitter
        should move on to the next byte after the given cycle

    rx_data: Signal(8), output
        data received from the PHY; valid when rx_valid is asserted
    rx_valid: Signal(), output
        indicates that the data present on rx_data is new and valid data; goes high for a single usb
        clock cycle to indicate new data is ready
    rx_active: Signal(), output
        indicates that the PHY is actively receiving data from the host; data is only valid when
        :attr:``rx_valid`` is high
    rx_error: Signal(), output
        indicates that an error has occurred in the current transmission

    rx_complete: Signal(), output:
        strobe that goes high for one cycle when a packet rx is complete


    line_state: Signal(2), output
        Indicates the current state of the D+ and D- lines. Matches the UTMI specification values,
        where 0 = SE0, 1=K, and 2=J.
    vbus_valid: Signal(), output
        Indicates that a valid VBUS signal is present. This signal is valid iff the I/O parameter
        contains a ``vbus_valid`` element; otherwise it is hard connected to '1'
    session_end: Signal(), output
        Indicates that no VBUS signal is present. This signal is valid iff the I/O parameter
        contains a ``vbus_valid`` element; otherwise it is hard connected to '0'

    xcvr_select: Signal(2), input
        Selects the active USB speed. This transceiver only functions as a full speed transciever; so
        this signal is effectively ignored. To support connection to high-speed gateware, this module
        will prevent the USB lines from being driven when this signal is 0b00; allowing the gateware to
        attempt a high-speed detection handshake without adverse affect.
    term_select: Signal(), input
        When asserted, this will connect the device's full speed pull-up resistor.
    op_mode: Signal(2), input
        Selects the operating mode of the UTMI transceiver. A value of 0 causes normal operations;
        a value of 1 prevents D+ and D- from being driven; and a value of 2 disables bit-stuffing.

    dm_pulldown: Signal(), input
        When asserted, this will indicate that the host-mode's D- pulldown should be connected.
    dm_pulldown: Signal(), input
        When asserted, this will indicate that the host-mode's D+ pulldown should be connected.

    Parameters
    -----------

    io: Record(d_p, d_n, [pullup], [pulldown], [vbus_valid])
        A record containing the raw I/O signals to be used to drive our I/O-based USB connnection.
        The ``d_p`` and ``d_n`` signals are mandatory; the ``pullup``, ``pulldown``,
        and ``vbus_valid`` signals are optional.
    """

    OP_MODE_NORMAL      = 0b00
    OP_MODE_NONDRIVING  = 0b10
    OP_MODE_NO_ENCODING = 0b01

    def __init__(self, *, io):
        self._io = io

        #
        # I/O port
        #
        self.tx_data     = Signal(8)
        self.tx_valid    = Signal()
        self.tx_ready    = Signal()

        self.rx_data     = Signal(8)
        self.rx_valid    = Signal()
        self.rx_active   = Signal()
        self.rx_error    = Signal()
        self.rx_complete = Signal()

        self.line_state  = Signal(2)
        self.vbus_valid  = Signal()
        self.session_end = Signal()

        self.xcvr_select = Signal(2)
        self.term_select = Signal()
        self.op_mode     = Signal(2)

        self.dp_pulldown = Signal()
        self.dm_pulldown = Signal()


    def elaborate(self, platform):
        m = Module()


        #
        # General state signals.
        #

        # Our line state is always taken directly from D- and D+.
        m.d.comb += self.line_state.eq(Cat(self._io.d_n.i, self._io.d_p.i))

        # If we have a ``vbus_valid`` indication, use it to drive our ``vbus_valid``
        # signal. Otherwise, we'll pretend ``vbus_valid`` is always true, for compatibility.
        if hasattr(self._io, 'vbus_valid'):
            m.d.comb += [
                self.vbus_valid   .eq(self._io.vbus_valid.i),
                self.session_end  .eq(~self._io.vbus_valid.i)
            ]
        else:
            m.d.comb += [
                self.vbus_valid   .eq(1),
                self.session_end  .eq(0)
            ]


        #
        # General control signals.
        #

        # If we have a pullup signal, drive it based on ``term_select``.
        if hasattr(self._io, 'pullup'):
            m.d.comb += self._io.pullup.o.eq(self.term_select)

        # If we have a pulldown signal, drive it based on our pulldown controls.
        if hasattr(self._io, 'pulldown'):
            m.d.comb += self._io.pullup.o.eq(self.dm_pulldown | self.dp_pulldown)


        #
        # Transmitter
        #
        in_normal_mode       = (self.op_mode == self.OP_MODE_NORMAL)
        in_non_encoding_mode = (self.op_mode == self.OP_MODE_NO_ENCODING)

        m.submodules.transmitter = transmitter = TxPipeline()


        # When we're in normal mode, we'll drive the USB bus with our standard USB transmitter data.
        with m.If(in_normal_mode):
            m.d.comb += [

                # UTMI Transmit data.
                transmitter.i_data_payload  .eq(self.tx_data),
                transmitter.i_oe            .eq(self.tx_valid),
                self.tx_ready               .eq(transmitter.o_data_strobe),

                # USB output.
                self._io.d_p.o   .eq(transmitter.o_usbp),
                self._io.d_n.o   .eq(transmitter.o_usbn),

                # USB tri-state control.
                self._io.d_p.oe  .eq(transmitter.o_oe),
                self._io.d_n.oe  .eq(transmitter.o_oe),
            ]


        # When we're in non-encoding mode ("disable bitstuff and NRZI"),
        # we'll output to D+ and D- directly when tx_valid is true.
        with m.Elif(in_non_encoding_mode):
            m.d.comb += [
                # USB output.
                self._io.d_p.o   .eq(self.tx_data),
                self._io.d_n.o   .eq(~self.tx_data),

                # USB tri-state control.
                self._io.d_p.oe  .eq(self.tx_valid),
                self._io.d_n.oe  .eq(self.tx_valid)
            ]

        # When we're in other modes (non-driving or invalid), we'll not output at all.
        # This block does nothing, as signals default to zero, but it makes the intention obvious.
        with m.Else():
            m.d.comb += [
                self._io.d_p.oe  .eq(0),
                self._io.d_n.oe  .eq(0),
            ]

        # Generate our USB clock strobe, which should pulse at 12MHz.
        counter = Signal(2)
        m.d.usb_io += counter.eq(counter + 1)
        m.d.comb += transmitter.i_bit_strobe.eq(counter == 0)

        #
        # Receiver
        #
        m.submodules.receiver = receiver = RxPipeline()
        m.d.comb += [

            # We'll listen for packets on D+ and D- _whenever we're not transmitting._.
            # (If we listen while we're transmitting, we'll hear our own packets.)
            receiver.i_usbp  .eq(self._io.d_p.i & ~transmitter.o_oe),
            receiver.i_usbn  .eq(self._io.d_n.i & ~transmitter.o_oe),

            self.rx_data     .eq(receiver.o_data_payload),
            self.rx_valid    .eq(receiver.o_data_strobe & receiver.o_pkt_in_progress),
            self.rx_active   .eq(receiver.o_pkt_in_progress),
            self.rx_error    .eq(receiver.o_receive_error)
        ]
        m.d.usb += self.rx_complete .eq(receiver.o_pkt_end)


        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Utilities for building USB3 descriptors into gateware. """

from amaranth                          import *
from usb_protocol.emitters.descriptors import DeviceDescriptorCollection

from ...stream                         import SuperSpeedStreamInterface
from ....stream.generator              import ConstantStreamGenerator


class GetDescriptorHandler(Elaboratable):
    """ Gateware that handles responding to GetDescriptor requests.

    Currently does not support descriptors in multiple languages.

    Attributes
    ----------
    value: Signal(16), input
        The value field associated with the Get Descriptor request. Contains the descriptor type and index.
    length: Signal(16), input
        The length field associated with the Get Descriptor request.
        Determines the maximum amount allowed in a response.

    start: Signal(), input
        Strobe that indicates when a descriptor should be transmitted.

    tx: stream_type(), output stream
        Stream that carries our output descriptor data.
    tx_length: Signal(16), output
        The actual length of the descriptor to be sent. At SuperSpeed, used for header generation.

    stall: Signal(), output
        Strobe; pulsed if a STALL handshake should be generated, instead of a response.

    Parameters
    ----------
    domain: string
        The clock domain this generator should belong to. Defaults to ''.
    stream_type: StreamInterface, or subclass
        The type of stream we'll be multiplexing.
    """

    def __init__(self, descriptor_collection: DeviceDescriptorCollection, *,
        usb_domain="ss", stream_type=SuperSpeedStreamInterface):
        """
        Parameters
        ----------
        descriptor_collection: DeviceDescriptorCollection
            The DeviceDescriptorCollection containing the descriptors to use for this device.
        usb_domain: string
            The name of the domain to use for USB data exchange.
        stream_type: StreamInterface subclass
            The type of the stream to be used to carry descriptor data.
        """
        self._domain      = usb_domain
        self._descriptors = descriptor_collection
        self._stream_type = stream_type

        #
        # I/O port
        #
        self.value     = Signal(16)
        self.length    = Signal(16)

        self.start     = Signal()

        self.tx        = stream_type()
        self.tx_length = Signal(16)
        self.stall     = Signal()


    def elaborate(self, platform):
        m = Module()

        # Collection that will store each of our descriptor-generation submodules.
        descriptor_generators = {}

        #
        # Create our constant-stream generators for each of our descriptors.
        #
        for type_number, index, raw_descriptor in self._descriptors:

            # Create the generator...
            generator = ConstantStreamGenerator(raw_descriptor,
                domain=self._domain, stream_type=self._stream_type, max_length_width=16)
            descriptor_generators[(type_number, index)] = generator

            # ... and attach it to this module.
            m.submodules += generator


        #
        # Connect up each of our generators.
        #

        with m.Switch(self.value):

            # Generate a conditional interconnect for each of our items.
            for (type_number, index), generator in descriptor_generators.items():

                # If the value matches the given type number...
                with m.Case(type_number << 8 | index):

                    # ... connect the relevant generator to our output.
                    m.d.comb += [
                        generator.start       .eq(self.start),
                        generator.max_length  .eq(self.length),
                    ]

                    # Buffer the output stream to improve timings.
                    with m.If(~self.tx.valid.any() | self.tx.ready):
                        m.d.sync += [
                            self.tx               .stream_eq(generator.stream, omit={'ready'}),
                            self.tx_length        .eq(generator.output_length)
                        ]
                        m.d.comb += [
                            generator.stream.ready.eq(1),
                        ]

            # If none of our descriptors match, stall any request that comes in.
            with m.Default():
                m.d.comb += self.stall.eq(self.start)

        # Convert our sync domain to the domain requested by the user, if necessary.
        if self._domain != "sync":
            m = DomainRenamer({"sync": self._domain})(m)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Isochronous Timestamp Packet (ITP)-related gateware. """

from amaranth import *
from usb_protocol.types.superspeed import HeaderPacketType

from ..link.header import HeaderQueue, HeaderPacket


class TimestampPacketReceiver(Elaboratable):
    """ Gateware that receives Isochronous Timestamp Packets, and keeps time.

    Attributes
    ----------
    header_sink: HeaderQueue(), input stream
        Input stream carrying header packets from the link layer.

    bus_interval_counter: Signal(14), output
        The currently timestamp, expressed in a number of 125uS bus intervals.
    delta: Signal(13)
        The delta from the aligned bus interval and ITP transmission.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.header_sink          = HeaderQueue()

        self.update_received      = Signal()
        self.bus_interval_counter = Signal()
        self.delta                = Signal()


    def elaborate(self, platform):
        m = Module()

        # Accept any Isochronous Timestamp Packet...
        new_packet = self.header_sink.valid
        is_for_us  = self.header_sink.get_type() == HeaderPacketType.ISOCHRONOUS_TIMESTAMP
        with m.If(new_packet & is_for_us):
            m.d.comb += self.header_sink.ready.eq(1)

            # ... and extract its fields.
            packet = self.header_sink.header
            m.d.ss += [
                self.update_received       .eq(1),
                self.bus_interval_counter  .eq(packet.dw0[ 5:19]),
                self.delta                 .eq(packet.dw0[19:32])
            ]
        with m.Else():
            m.d.ss += self.update_received.eq(0)

        return m



# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Helpers for clock domain crossings. """

from amaranth       import Signal


def stretch_strobe_signal(m, strobe, *, to_cycles, output=None, domain=None, allow_delay=False):
    """ Stretches a given strobe to the given number of cycles.

    Parameters:
        strobe    -- The strobe signal to stretch.
        to_cycles -- The number of cycles to stretch the given strobe to. Must be >= 1.

        output    -- If provided, the given signal will be used as the output signal.
        domain    -- If provided, the given domain _object_ will be used in lieu of the sync domain.

     Returns the output signal. If output is provided, this is the same signal; otherwise, it is the
     signal that was created internally.
     """

    # Assume the sync domain if no domain is provided.
    if domain is None:
        domain = m.d.sync

    # If we're not given an output signal to target, create one.
    if output is None:
        output = Signal()

    # Special case: if to_cycles is '1', we don't need to modify the strobe.
    # Connect it through directly.
    if to_cycles == 1:
        m.d.comb += output.eq(strobe)
        return output

    # Create a signal that shifts in our strobe constantly, so we
    # have a memory of its last N values.
    if allow_delay:
        delayed_strobe = Signal(to_cycles)
        domain += delayed_strobe.eq((delayed_strobe << 1) | strobe)
        m.d.comb += output.eq(delayed_strobe != 0)
    else:
        delayed_strobe = Signal(to_cycles - 1)
        domain += delayed_strobe.eq((delayed_strobe << 1) | strobe)
        m.d.comb += output.eq(strobe | (delayed_strobe != 0))

    return output



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test.utils import LunaGatewareTestCase, sync_test_case

from luna.gateware.interface.uart import UARTMultibyteTransmitter, UARTTransmitter

class UARTTransmitterTest(LunaGatewareTestCase):
    DIVISOR = 10

    FRAGMENT_UNDER_TEST = UARTTransmitter
    FRAGMENT_ARGUMENTS = dict(divisor=DIVISOR)


    def advance_half_bit(self):
        yield from self.advance_cycles(self.DIVISOR // 2)

    def advance_bit(self):
        yield from self.advance_cycles(self.DIVISOR)


    def assert_data_sent(self, byte_expected):
        dut = self.dut

        # Our start bit should remain present until the next bit period.
        yield from self.advance_half_bit()
        self.assertEqual((yield dut.tx), 0)

        # We should then see each bit of our data, LSB first.
        bits = [int(i) for i in f"{byte_expected:08b}"]
        for bit in bits[::-1]:
            yield from self.advance_bit()
            self.assertEqual((yield dut.tx), bit)

        # Finally, we should see a stop bit.
        yield from self.advance_bit()
        self.assertEqual((yield dut.tx), 1)


    @sync_test_case
    def test_burst_transmit(self):
        dut = self.dut
        stream = dut.stream

        # We should remain idle until a transmit is requested...
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.idle), 1)
        self.assertEqual((yield dut.stream.ready), 1)

        # ... and our tx line should idle high.
        self.assertEqual((yield dut.tx), 1)

        # First, transmit 0x55 (maximum transition rate).
        yield stream.payload.eq(0x55)
        yield stream.valid.eq(1)

        # We should see our data become accepted; and we
        # should see a start bit.
        yield
        self.assertEqual((yield stream.ready), 1)
        yield
        self.assertEqual((yield dut.tx), 0)

        # Provide our next byte of data once the current
        # one has been accepted. Changing this before the tests
        # below ensures that we validate that data is latched properly.
        yield stream.payload.eq(0x66)

        # Ensure we get our data correctly.
        yield from self.assert_data_sent(0x55)
        yield from self.assert_data_sent(0x66)

        # Stop transmitting after the next frame.
        yield stream.valid.eq(0)

        # Ensure we actually stop.
        yield from self.advance_bit()
        self.assertEqual((yield dut.idle), 1)


class UARTMultibyteTransmitterTest(LunaGatewareTestCase):
    DIVISOR = 10

    FRAGMENT_UNDER_TEST = UARTMultibyteTransmitter
    FRAGMENT_ARGUMENTS = dict(divisor=DIVISOR, byte_width=4)


    def advance_half_bit(self):
        yield from self.advance_cycles(self.DIVISOR // 2)

    def advance_bit(self):
        yield from self.advance_cycles(self.DIVISOR)


    def assert_data_sent(self, byte_expected):
        dut = self.dut

        # Our start bit should remain present until the next bit period.
        yield from self.advance_half_bit()
        self.assertEqual((yield dut.tx), 0)

        # We should then see each bit of our data, LSB first.
        bits = [int(i) for i in f"{byte_expected:08b}"]
        for bit in bits[::-1]:
            yield from self.advance_bit()
            self.assertEqual((yield dut.tx), bit)

        # Finally, we should see a stop bit.
        yield from self.advance_bit()
        self.assertEqual((yield dut.tx), 1)

        yield from self.advance_cycles(2)


    @sync_test_case
    def test_burst_transmit(self):
        dut = self.dut
        stream = dut.stream

        # We should remain idle until a transmit is requested...
        yield from self.advance_cycles(10)
        self.assertEqual((yield dut.idle), 1)
        self.assertEqual((yield dut.stream.ready), 1)

        # Transmit a four-byte word.
        yield stream.payload.eq(0x11223355)
        yield stream.valid.eq(1)

        # We should see our data become accepted; and we
        # should see a start bit.
        yield
        self.assertEqual((yield stream.ready), 1)

        # Ensure we get our data correctly, and that our transmitter
        # isn't accepting data mid-frame.
        yield from self.assert_data_sent(0x55)
        self.assertEqual((yield stream.ready), 0)
        yield from self.assert_data_sent(0x33)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

"""
This module contains definitions of memory units that work well for USB applications.
"""

from amaranth import Elaboratable, Module, Signal
from amaranth.lib.memory import Memory
from amaranth.hdl.xfrm import DomainRenamer


class TransactionalizedFIFO(Elaboratable):
    """ Transactionalized, buffer first-in-first-out queue.

    This FIFO is "transactionalized", which means that it allows sets of reads and writes to be "undone".
    Effectively, this FIFO allows "rewinding" its read and write pointers to a previous point in time,
    which makes it ideal for USB transmission or receipt; where the protocol can require blocks of data
    to be retransmitted or ignored.

    Attributes
    ----------
    read_data: Signal(width), output
        Contains the next byte in the FIFO. Valid only when :attr:``empty`` is false.
    read_en: Signal(), input
        When asserted, the current :attr:``read_data`` will move to the next value. The data is not
        internally consumed/dequeued until :attr:``read_commit`` is asserted. This read can be "undone"
        by asserting :attr:``read_discard``. Should only be asserted when :attr:``empty`` is false.
    read_commit: Signal(), input
        Strobe; when asserted, any reads performed since the last commit will be "finalized".
        This effectively frees the memory associated with past reads. If this value is tied to '1',
        the read port on this FIFO gracefully degrades to non-transactionalized port.
    read_discard: Signal(), input
        Strobe; when asserted; any reads since the last commit will be "undone", placing the read pointer
        back at the queue position it had after the last :attr:``read_commit`.
    empty: Signal(), output
        Asserted when no data is available in the FIFO. This signal refers to whether data is available to
        read. :attr:``read_commit`` will not change this value; but :attr:``read_discard`` will.


    write_data: Signal(width), input
        Holds the byte to be added to the FIFO when :attr:``write_en`` is asserted.
    write_en: Signal(), input
        When asserted, the current :attr:``write_data`` will be added to the FIFO; but will not be ready for read
        until :attr:``write_commit`` is asserted. This write can be "undone" by asserting :attr:``write_discard``.
        Should only be asserted when :attr:``full`` is false.
    write_commit: Signal(), input
        Strobe; when asserted, any writes reads performed since the last commit will be "finalized".
        This makes the relevant data available for read.
    write_discard: Signal(), input
        Strobe; when asserted; any writes since the last commit will be "undone", placing the write pointer
        back at the queue position it had after the last :attr:``write_commit`. This frees the relevant memory
        for new writes.
    full: Signal(), output
        Asserted when no space is available for writes in the FIFO. :attr:``write_commit`` will not change
        this value; but :attr:``write_discard`` will.

    space_available: Signal(range(0, depth + 1)), output
        Indicates the amount of space available in the FIFO. Useful for knowing whether we can add e.g. an
        entire packet to the FIFO.


    Attributes
    ----------
    width: int
        The width of each entry in the FIFO.
    depth: int
        The number of allowed entries in the FIFO.
    name: str
        The name of the relevant FIFO; to produce nicer debug output.
        If not provided, Amaranth will attempt auto-detection.
    domain: str
        The name of the domain this module should exist in.
    """

    def __init__(self, *, width, depth, name=None, domain="sync"):
        self.width  = width
        self.depth  = depth
        self.name   = name
        self.domain = domain

        #
        # I/O port
        #
        self.read_data        = Signal(width)
        self.read_en          = Signal()
        self.read_commit      = Signal()
        self.read_discard     = Signal()
        self.empty            = Signal()

        self.write_data       = Signal(width)
        self.write_en         = Signal()
        self.write_commit     = Signal()
        self.write_discard    = Signal()
        self.full             = Signal()

        self.space_available  = Signal(range(0, depth + 1))


    def elaborate(self, platform):
        m = Module()

        # Range shortcuts for internal signals.
        address_range = range(0, self.depth + 1)

        #
        # Core internal "backing store".
        #
        m.submodules[self.name] = memory = Memory(shape=self.width, depth=self.depth + 1, init=[])
        read_port = memory.read_port()
        write_port = memory.write_port()

        # Always connect up our memory's data/en ports to ours.
        m.d.comb += [
            self.read_data  .eq(read_port.data),

            write_port.data .eq(self.write_data),
            write_port.en   .eq(self.write_en & ~self.full)
        ]

        #
        # Write port.
        #

        # We'll track two pieces of data: our _committed_ write position, and our current un-committed write one.
        # This will allow us to rapidly backtrack to our pre-commit position.
        committed_write_pointer = Signal(address_range)
        current_write_pointer   = Signal(address_range)
        m.d.comb += write_port.addr.eq(current_write_pointer)


        # Compute the location for the next write, accounting for wraparound. We'll not assume a binary-sized
        # buffer; so we'll compute the wraparound manually.
        next_write_pointer      = Signal.like(current_write_pointer)
        with m.If(current_write_pointer == self.depth):
            m.d.comb += next_write_pointer.eq(0)
        with m.Else():
            m.d.comb += next_write_pointer.eq(current_write_pointer + 1)


        # If we're writing to the fifo, update our current write position.
        with m.If(self.write_en & ~self.full):
            m.d.sync += current_write_pointer.eq(next_write_pointer)

        # If we're committing a FIFO write, update our committed position.
        with m.If(self.write_commit):
            m.d.sync += committed_write_pointer.eq(current_write_pointer)

        # If we're discarding our current write, reset our current position,
        with m.If(self.write_discard):
            m.d.sync += current_write_pointer.eq(committed_write_pointer)


        #
        # Read port.
        #

        # We'll track two pieces of data: our _committed_ read position, and our current un-committed read one.
        # This will allow us to rapidly backtrack to our pre-commit position.
        committed_read_pointer = Signal(address_range)
        current_read_pointer   = Signal(address_range)


        # Compute the location for the next read, accounting for wraparound. We'll not assume a binary-sized
        # buffer; so we'll compute the wraparound manually.
        next_read_pointer      = Signal.like(current_read_pointer)
        with m.If(current_read_pointer == self.depth):
            m.d.comb += next_read_pointer.eq(0)
        with m.Else():
            m.d.comb += next_read_pointer.eq(current_read_pointer + 1)


        # Our memory always takes a single cycle to provide its read output; so we'll update its address
        # "one cycle in advance". Accordingly, if we're about to advance the FIFO, we'll use the next read
        # address as our input. If we're not, we'll use the current one.
        with m.If(self.read_en & ~self.empty):
            m.d.comb += read_port.addr.eq(next_read_pointer)
        with m.Else():
            m.d.comb += read_port.addr.eq(current_read_pointer)


        # If we're reading from our the fifo, update our current read position.
        with m.If(self.read_en & ~self.empty):
            m.d.sync += current_read_pointer.eq(next_read_pointer)

        # If we're committing a FIFO write, update our committed position.
        with m.If(self.read_commit):
            m.d.sync += committed_read_pointer.eq(current_read_pointer)

        # If we're discarding our current write, reset our current position,
        with m.If(self.read_discard):
            m.d.sync += current_read_pointer.eq(committed_read_pointer)


        #
        # FIFO status.
        #

        # Our FIFO is empty if our read and write pointers are in the same. We'll use the current
        # read position (which leads ahead) and the committed write position (which lags behind).
        m.d.comb += self.empty.eq(current_read_pointer == committed_write_pointer)

        # For our space available, we'll use the current write position (which leads ahead) and our committed
        # read position (which lags behind). This yields two cases: one where the buffer isn't wrapped around,
        # and one where it is.
        with m.If(self.full):
            m.d.comb += self.space_available.eq(0)
        with m.Elif(committed_read_pointer <= current_write_pointer):
            m.d.comb += self.space_available.eq(self.depth - (current_write_pointer - committed_read_pointer))
        with m.Else():
            m.d.comb += self.space_available.eq(committed_read_pointer - current_write_pointer - 1)

        # Our FIFO is full if we don't have any space available.
        m.d.comb += self.full.eq(next_write_pointer == committed_read_pointer)


        # If we're not supposed to be in the sync domain, rename our sync domain to the target.
        if self.domain != "sync":
            m = DomainRenamer({"sync": self.domain})(m)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Pure-gateware, UTMI-compatible Full Speed PHY."""

from .phy import GatewarePHY



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Packet transmission handling gateware. """

from amaranth                      import *
from usb_protocol.types.superspeed import LinkCommand, HeaderPacketType

from .header                       import HeaderPacket, HeaderQueue
from .crc                          import compute_usb_crc5, HeaderPacketCRC, DataPacketPayloadCRC
from .command                      import LinkCommandDetector
from ..physical.coding             import SHP, SDP, EPF, END, EDB, get_word_for_symbols
from ...stream                     import USBRawSuperSpeedStream, SuperSpeedStreamInterface



class RawPacketTransmitter(Elaboratable):
    """ Class that generates and sends header packets; with an optional payload attached.

    This class generates the checks required at the link layer of the USB specification;
    which include checking the CRC-5 and CRC-16 embedded within the header packet, and
    the CRC-32 embedded within any associated payloads.


    Attributes
    ----------
    source: USBRawSuperSpeedStream(), output stream
        Stream that carries the generated packets.

    header: HeaderPacket(), input
        The header packet to be sent. The CRC fields do not need to be valid, and will
        be ignored and replaced with a computed CRC. If the header packet is of type data,
        :attr:``data_sink`` will be allowed to provide a follow-up Data Packet Payload.
    data_sink: SuperSpeedStreamInterface(), input stream
        If the provided :attr:``header`` is a Data Header, and :attr:``data_sink`` is valid
        when the packet's transmission is sent, this data stream will be used to provide a
        follow-up Data Packet Payload.

    generate: Signal(), input
        Strobe; indicates that a packet should be generated.
    done: Signal(), output
        Indicates that the packet will be complete this cycle; and thus this unit will
        be ready to send another packet next cycle.
    """

    def __init__(self):

        #
        # I/O port.
        #
        self.source    = USBRawSuperSpeedStream()

        # Packet data in.
        self.header    = HeaderPacket()
        self.data_sink = SuperSpeedStreamInterface()

        self.generate  = Signal()
        self.done      = Signal()


    def elaborate(self, platform):
        m = Module()

        # Shorthands.
        source    = self.source
        data_sink = self.data_sink

        # Latched data.
        header    = HeaderPacket()

        #
        # CRC Generators
        #

        # CRC-16, for header packets
        m.submodules.crc16 = crc16 = HeaderPacketCRC()
        m.d.comb += crc16.data_input.eq(source.data)

        # CRC-32, for payload packets
        m.submodules.crc32 = crc32 = DataPacketPayloadCRC()
        m.d.comb += [
            crc32.data_input    .eq(self.data_sink.data),

            # Advance our CRC32 whenever we accept data from our sink,
            # according to how many bytes are currently valid.
            crc32.advance_word  .eq((data_sink.valid == 0b1111) & data_sink.ready),
            crc32.advance_3B    .eq((data_sink.valid == 0b0111) & data_sink.ready),
            crc32.advance_2B    .eq((data_sink.valid == 0b0011) & data_sink.ready),
            crc32.advance_1B    .eq((data_sink.valid == 0b0001) & data_sink.ready),
        ]


        #
        # Packet transmitter.
        #

        # Create a time-delayed version of our data; so we can always work with data from
        # a single cycle ago. This allows us to compute our CRC before we need to send it;
        # which is important, as a CRC can start in the middle of a word (and thus needs to
        # factor in data available the same cycle as the CRC is sent).
        pipelined_data_word  = Signal.like(self.data_sink.data)
        pipelined_data_valid = Signal.like(self.data_sink.valid)

        # Store whether our packet is a ZLP.
        packet_is_zlp = Signal()

        with m.FSM(domain="ss"):

            # IDLE -- wait for a generate command
            with m.State("IDLE"):
                # Don't start our CRCs until we're sending the data section relevant to them.
                m.d.comb += [
                    crc16.clear  .eq(1),
                    crc32.clear  .eq(1)
                ]

                # Once we have a request, latch in our data, and start sending.
                with m.If(self.generate):
                    m.d.ss += header.eq(self.header)
                    m.next = "SEND_HPSTART"


            # SEND_HPSTART -- send our "start-of-header-packet" marker.
            with m.State("SEND_HPSTART"):

                # Drive the bus with our header...
                header_data, header_ctrl = get_word_for_symbols(SHP, SHP, SHP, EPF)
                m.d.comb += [
                    source.valid  .eq(1),
                    source.data   .eq(header_data),
                    source.ctrl   .eq(header_ctrl),
                ]

                # ... and keep driving it until it's accepted.
                with m.If(source.ready):
                    m.next = "SEND_DW0"


            # SEND_DWn -- send along the three first data words unalatered, as we don't
            # need to fill in any CRC details, here.
            data_words = [header.dw0, header.dw1, header.dw2]
            for n in range(3):
                with m.State(f"SEND_DW{n}"):

                    # Drive the bus with the relevant data word...
                    m.d.comb += [
                        source.valid  .eq(1),
                        source.data   .eq(data_words[n]),
                        source.ctrl   .eq(0),
                    ]

                    # ... and keep driving it until it's accepted.
                    with m.If(source.ready):
                        m.d.comb += crc16.advance_crc.eq(1)
                        m.next = f"SEND_DW{n + 1}"


            # Compose our final data word from our individual fields, filling
            with m.State("SEND_DW3"):

                # Compose our data word
                m.d.comb += [
                    source.valid        .eq(1),
                    source.data[ 0:16]  .eq(crc16.crc),
                    source.data[16:19]  .eq(header.sequence_number),
                    source.data[19:22]  .eq(header.dw3_reserved),
                    source.data[22:25]  .eq(header.hub_depth),
                    source.data[25]     .eq(header.delayed),
                    source.data[26]     .eq(header.deferred),
                    source.data[27:32]  .eq(compute_usb_crc5(source.data[16:27])),
                    source.ctrl         .eq(0)
                ]

                # Once our final header packet word has been sent; decide what to do next.
                with m.If(source.ready):

                    # If we just sent a data packet header, and we have a valid data-stream,
                    # or if we're sending a ZLP, follow on immediately with a Data Packet Payload.
                    was_data_header = (header.dw0[ 0: 4] == HeaderPacketType.DATA)
                    with m.If(was_data_header):
                        m.d.ss += packet_is_zlp.eq(self.data_sink.valid == 0)
                        m.next = "START_DPP"

                    # Otherwise, we're done!
                    with m.Else():
                        m.d.comb += self.done.eq(1)
                        m.next = f"IDLE"


            # START_DPP -- we'll start our data packet payload with our framing;
            # and prepare to send our actual payload.
            with m.State("START_DPP"):

                # Send our start framing...
                header_data, header_ctrl = get_word_for_symbols(SDP, SDP, SDP, EPF)
                m.d.comb += [
                    source.valid  .eq(1),
                    source.data   .eq(header_data),
                    source.ctrl   .eq(header_ctrl),
                ]

                with m.If(source.ready):

                    # Special case: if we're retransmitting a data packet, we'll abort it immediately,
                    # since we don't have the data buffered anywhere. Retransmission of the data payload
                    # will be handled at the protocol layer.
                    with m.If(header.delayed):
                        m.next = "ABORT_DPP"

                    # Special case: if we're sending a ZLP, we'll jump directly to sending our CRC.
                    with m.Elif(packet_is_zlp):

                        # We'll treat this as though we'd just sent a full data word; as this indicates
                        # to our later states that our data was word-aligned. (0B is a multiple of 4, after all.)
                        m.d.ss += pipelined_data_valid.eq(0b1111)
                        m.next = "SEND_CRC"

                    # Otherwise, we have some data to handle.
                    with m.Else():

                        # Capture the first word of our data-to-send into our pipeline...
                        m.d.ss   += [
                            pipelined_data_word   .eq(self.data_sink.data),
                            pipelined_data_valid  .eq(self.data_sink.valid)
                        ]

                        # ... and accept that data, so our sender moves on to the next word.
                        m.d.comb += self.data_sink.ready.eq(1)

                        # ... and as long as we didn't just capture the last word, start the actual
                        # payload transmission.
                        with m.If(~self.data_sink.last):
                            m.next = "SEND_PAYLOAD"

                        # If we -did- just capture the last word, we can skip to our last-word handler.
                        with m.Else():
                            m.next = "SEND_LAST_WORD"


            # SEND_PAYLOAD -- send the core part of our data stream.
            with m.State("SEND_PAYLOAD"):

                # We'll always send our last pipelined value; so we can always have a valid "last word"
                # for our last word handler, which handles special cases like inserting our CRC.
                m.d.comb += [
                    source.data         .eq(pipelined_data_word),
                    source.valid        .eq(1),
                ]

                # Each time one of the words we're transmitting is accepted, we'll in turn accept a new
                # word into our pipeline register.
                with m.If(source.ready):
                    m.d.comb += self.data_sink.ready.eq(1)
                    m.d.ss   += [
                        pipelined_data_word   .eq(self.data_sink.data),
                        pipelined_data_valid  .eq(self.data_sink.valid)
                    ]

                    # When we're finally receiving our last packet, we're also finished computing our CRC.
                    # This means we can move on safely to sending our pipelined word; knowing we're ready to
                    # send part of our CRC if the last word isn't fully aligned.
                    with m.If(data_sink.last):
                        m.next = "SEND_LAST_WORD"


            # SEND_LAST_WORD -- flush our pipeline; and send our final word.
            # Our final word is a special case; as it may not be a full word; in which
            # case we'll have to stick part of our CRC into it.
            with m.State("SEND_LAST_WORD"):

                # Send the final value captured from our pipeline; which we may modify slightly below.
                m.d.comb += [
                    source.data         .eq(pipelined_data_word),
                    source.valid        .eq(1),
                ]

                # If we didn't capture a full word, we'll need to stick pieces of our final CRC
                # into the unused/invalid sections of our data. We'll do so, selecting our position
                # based on how many bytes of valid data we're sending.
                with m.Switch(pipelined_data_valid):

                    # 3 valid bytes, 1 byte of CRC
                    with m.Case (0b0111):
                        m.d.comb += source.data[24:32].eq(crc32.crc[0:8])

                    # 2 valid bytes, 2 bytes of CRC
                    with m.Case (0b0011):
                        m.d.comb += source.data[16:32].eq(crc32.crc[0:16])

                    # 1 valid byte, 3 bytes of CRC
                    with m.Case (0b0001):
                        m.d.comb += source.data[8:32].eq(crc32.crc[0:24])


                # Once our last data word is accepted, move to sending the (remainder) of the CRC.
                with m.If(source.ready):
                    m.next = "SEND_CRC"


            # SEND_CRC -- send the CRC (or what remains of it)
            with m.State("SEND_CRC"):
                m.d.comb += source.valid.eq(1)

                # We'll need to send however much of the CRC is left; which will depend on the validity
                # of the final byte we had pipelined.
                with m.Switch(pipelined_data_valid):

                    # If our data packet was word aligned, we've sent no CRC bytes.
                    # Send our full word.
                    with m.Case(0b1111):
                        m.d.comb += [
                            source.data  .eq(crc32.crc),
                            source.ctrl  .eq(0)
                        ]

                    # If we had three valid bytes of data last time, we've sent one byte of CRC.
                    # Send the other three, followed by one END framing word.
                    with m.Case(0b0111):
                        m.d.comb += [
                            source.data  .eq(Cat(crc32.crc[8:32], END.value_const())),
                            source.ctrl  .eq(0b1000),
                        ]

                    # Same, but for 2B valid and 2B CRC
                    with m.Case(0b0011):
                        m.d.comb += [
                            source.data  .eq(Cat(crc32.crc[16:32], END.value_const(repeat=2))),
                            source.ctrl  .eq(0b1100),
                        ]

                    # Same, but for 1B valid and 3B CRC
                    with m.Case(0b0001):
                        m.d.comb += [
                            source.data  .eq(Cat(crc32.crc[24:32], END.value_const(repeat=3))),
                            source.ctrl  .eq(0b1110),
                        ]


                # Once our CRC is accepted, finish our payload.
                with m.If(source.ready):
                    m.next = "FINISH_DPP"


            # FINISH_DPP -- send our end-of-payload framing.
            with m.State("FINISH_DPP"):
                m.d.comb += source.valid.eq(1)

                # Get our data and ctrl constants for our framing.
                framing_data, framing_ctrl = get_word_for_symbols(END, END, END, EPF)

                # We'll need to send however much of the framing is left; which will depend on the
                # where our CRC ended.
                with m.Switch(pipelined_data_valid):

                    # If our data packet was word aligned, we've sent no CRC bytes.
                    # Send our full word.
                    with m.Case(0b1111):
                        m.d.comb += [
                            source.data         .eq(framing_data),
                            source.ctrl         .eq(framing_ctrl),
                        ]

                    # If we had three valid bytes of data last time, we've sent one byte of framing.
                    # Send the other three, followed by zeroes (logical IDL).
                    with m.Case(0b0111):
                        m.d.comb += [
                            source.data         .eq(framing_data[8:]),
                            source.ctrl         .eq(framing_ctrl[1:]),
                        ]

                    # Same, but 2B frame and 2B IDL.
                    with m.Case(0b0011):
                        m.d.comb += [
                            source.data         .eq(framing_data[16:]),
                            source.ctrl         .eq(framing_ctrl[ 2:]),
                        ]

                    # Same, but 1B frame and 3B IDL.
                    with m.Case(0b0001):
                        m.d.comb += [
                            source.data         .eq(framing_data[24:]),
                            source.ctrl         .eq(framing_ctrl[ 3:]),
                        ]


                # Once our CRC is accepted, finish our payload.
                with m.If(source.ready):
                    m.d.comb += self.done.eq(1)
                    m.next = "IDLE"


            # ABORT_DPP -- send our abort-of-payload framing.
            with m.State("ABORT_DPP"):

                # Send our abort framing...
                framing_data, framing_ctrl = get_word_for_symbols(EDB, EDB, EDB, EPF)
                m.d.comb += [
                    source.valid  .eq(1),
                    source.data   .eq(framing_data),
                    source.ctrl   .eq(framing_ctrl),
                ]

                # Once our framing is accepted, finish our payload.
                with m.If(source.ready):
                    m.d.comb += self.done.eq(1)
                    m.next = "IDLE"

        return m



class PacketTransmitter(Elaboratable):
    """ Transmitter-side Header Packet logic.

    This module handles all header-packet-transmission related logic for the link layer; including
    header packet generation, partner buffer tracking / flow control, and link command reception.

    Attributes
    ----------
    sink: USBRawSuperSpeedStream(), input stream [monitor only]
        Stream that monitors the data received on the physical layer for link commands.
    source: USBRawSuperSpeedStream(), output stream
        Stream that carries our transmissions down to the physical layer.

    enable: Signal(), input
        When asserted, this unit will be enabled; and will be allowed to start transmitting.

    queue: HeaderQueue(), input stream
        Stream of header packets received received from the protocol layer to be transmitted.

    retry_received: Signal(), output
        Strobe; pulsed high when we receive a link retry request.
    retry_required: Signal(), output
        Strobe; pulsed high when we need to send a retry request.

    lgo_received: Signal(), output
        Strobe; indicates that we've received a request to move to a new power state.
    lgo_target; Signal(2), output
        Indicates the power-state associated with a given LGO event; valid when :attr:``lgo_received``
        is asserted.

    recovery_required: Signal(), output
        Strobe; pulsed when a condition that requires link recovery occurs.
    """

    SEQUENCE_NUMBER_WIDTH = 3

    CREDIT_TIMEOUT = 5e-3

    def __init__(self, *, buffer_count=4, ss_clock_frequency=125e6):
        self._buffer_count    = buffer_count
        self._clock_frequency = ss_clock_frequency

        #
        # I/O port
        #
        self.sink                  = USBRawSuperSpeedStream()
        self.source                = USBRawSuperSpeedStream()

        # Simple controls.
        self.enable                = Signal()
        self.usb_reset             = Signal()
        self.bringup_complete      = Signal()

        # Protocol layer interface.
        self.queue                 = HeaderQueue()
        self.data_sink             = SuperSpeedStreamInterface()

        # Event interface.
        self.link_command_received = Signal()
        self.retry_received        = Signal()
        self.retry_required        = Signal()
        self.lrty_pending          = Signal()
        self.recovery_required     = Signal()

        self.lgo_received          = Signal()
        self.lgo_target            = Signal(2)

        # Debug information.
        self.credits_available     = Signal(range(self._buffer_count + 1))
        self.packets_to_send       = Signal(range(self._buffer_count + 1))


    def elaborate(self, platform):
        m = Module()

        #
        # Credit tracking.
        #

        # Keep track of how many transmit credits we currently have.
        credits_available    = Signal(range(self._buffer_count + 1))
        credit_received      = Signal()
        credit_consumed      = Signal()

        with m.If(credit_received & ~credit_consumed):
            m.d.ss += credits_available.eq(credits_available + 1)
        with m.Elif(credit_consumed & ~credit_received):
            m.d.ss += credits_available.eq(credits_available - 1)


        #
        # Task "queues".
        #

        # Control signals.
        retire_packet     = Signal()

        # Pending packet count.
        packets_to_send      = Signal(range(self._buffer_count + 1))
        enqueue_send         = Signal()
        dequeue_send         = Signal()

        # Un-retired packet count.
        packets_awaiting_ack = Signal(range(self._buffer_count + 1))

        # If we need to retry sending our packets, we'll need to reset our pending packet count.
        # Otherwise, we increment and decrement our "to send" counts normally.
        with m.If(self.retry_required):
            m.d.ss += packets_to_send.eq(packets_awaiting_ack)
        with m.Elif(enqueue_send & ~dequeue_send):
            m.d.ss += packets_to_send.eq(packets_to_send + 1)
        with m.Elif(dequeue_send & ~enqueue_send):
            m.d.ss += packets_to_send.eq(packets_to_send - 1)

        # Track how many packets are yet to be retired.
        with m.If(enqueue_send & ~retire_packet):
            m.d.ss += packets_awaiting_ack.eq(packets_awaiting_ack + 1)
        with m.Elif(retire_packet & ~enqueue_send & (packets_awaiting_ack != 0)):
            m.d.ss += packets_awaiting_ack.eq(packets_awaiting_ack - 1)


        #
        # Header Packet Buffers
        #

        # Track:
        # - which buffer should be filled next
        # - which buffer we should send from next
        # - which buffer we've last acknowledged
        read_pointer      = Signal(range(self._buffer_count))
        write_pointer     = Signal.like(read_pointer)
        ack_pointer       = Signal.like(read_pointer)

        # Create buffers to receive any incoming header packets.
        buffers = Array(HeaderPacket() for _ in range(self._buffer_count))

        # If we need to retry sending our packets, we'll need to start reading
        # again from the last acknowledged packet; so we'll reset our read pointer.
        with m.If(self.retry_required):
            m.d.ss += read_pointer.eq(ack_pointer)
        with m.Elif(dequeue_send):
            m.d.ss += read_pointer.eq(read_pointer + 1)

        # Last ACK'd buffer / packet retirement tracker.
        with m.If(retire_packet):
            m.d.ss += ack_pointer.eq(ack_pointer + 1)


        #
        # Packet acceptance (protocol layer -> link layer).
        #

        # Keep track of the next sequence number we'll need to assign.
        transmit_sequence_number = Signal(self.SEQUENCE_NUMBER_WIDTH)

        # If we have link credits available, we're able to accept data from the protocol layer.
        m.d.comb += self.queue.ready.eq(self.bringup_complete & (credits_available != 0))

        # If the protocol layer is handing us a packet...
        with m.If(self.queue.valid & self.queue.ready):
            # ... consume a credit, as we're going to be using up a receiver buffer, and
            # schedule sending of that packet.
            m.d.comb += [
                credit_consumed  .eq(1),
                enqueue_send     .eq(1)
            ]

            # Assign the packet a sequence number, and capture it into the buffer for transmission.
            # [USB3.0r1: 7.2.4.1.1]: "A header packet that is re-transmitted shall maintain its
            # originally assigned Header Sequence Number."
            m.d.ss += [
                buffers[write_pointer]                  .eq(self.queue.header),
                buffers[write_pointer].sequence_number  .eq(transmit_sequence_number),
                write_pointer                           .eq(write_pointer + 1),
                transmit_sequence_number                .eq(transmit_sequence_number + 1)
            ]


        #
        # Packet delivery (link layer -> physical layer)
        #
        m.submodules.packet_tx = packet_tx = RawPacketTransmitter()
        m.d.comb += [
            packet_tx.header     .eq(buffers[read_pointer]),
            packet_tx.data_sink  .stream_eq(self.data_sink),
            self.source          .stream_eq(packet_tx.source)
        ]


        # Keep track of whether a retry has been requested.
        retry_pending = Signal()
        with m.If(self.retry_required):
            m.d.ss += retry_pending.eq(1)


        with m.FSM(domain="ss"):

            # DISPATCH_PACKET -- wait packet transmissions to be scheduled, and prepare
            # our local transmitter with the proper data to send them.
            with m.State("DISPATCH_PACKET"):

                # If we have packets to send, pass them to our transmitter.
                with m.If(self.bringup_complete & (packets_to_send != 0)):

                    with m.If(~retry_pending):
                        # Wait until the packet is sent.
                        m.next = "WAIT_FOR_SEND"

                    with m.Else():
                        # Wait until all of the non-acknowledged packets are retransmitted.
                        m.next = "WAIT_FOR_RETRY"


            # WAIT_FOR_SEND -- we've now dispatched our packet; and we're ready to wait for it to be sent.
            with m.State("WAIT_FOR_SEND"):
                m.d.comb += packet_tx.generate.eq(1)

                # We're done with this packet.
                with m.If(packet_tx.done):

                    # If we received an LBAD in the meantime, our read pointer and counter are already
                    # set up for retransmission; don't touch them.
                    with m.If(~retry_pending):
                        m.d.comb += dequeue_send.eq(1)

                    # Handle the next packet, or wait for one.
                    m.next = "DISPATCH_PACKET"


            # WAIT_FOR_RETRY -- we're retransmitting all of the non-acknowledged packets, with the DL bit set;
            # but only after the receiver transmits LRTY.
            with m.State("WAIT_FOR_RETRY"):
                m.d.comb += packet_tx.header.delayed.eq(1)
                m.d.comb += packet_tx.generate.eq(~self.lrty_pending)

                # We're done with this packet.
                with m.If(packet_tx.done):
                    m.d.comb += dequeue_send.eq(1)

                    # If this was the last packet to retransmit, we're done handling this LBAD.
                    with m.If(packets_to_send == 1):
                        m.d.ss += retry_pending.eq(0)
                        m.next = "DISPATCH_PACKET"


        #
        # Link Command Handling
        #

        # Core link command receiver.
        m.submodules.lc_detector = lc_detector = LinkCommandDetector()
        m.d.comb += [
            lc_detector.sink            .tap(self.sink),
            self.link_command_received  .eq(lc_detector.new_command)
        ]

        # Keep track of which credit we expect to get back next.
        next_expected_credit = Signal(range(self._buffer_count))

        # Keep track of what sequence number we expect to have ACK'd next.
        next_expected_ack_number = Signal(self.SEQUENCE_NUMBER_WIDTH, init=-1)

        # Handle link commands as we receive them.
        with m.If(lc_detector.new_command):
            with m.Switch(lc_detector.command):

                #
                # Link Credit Reception
                #
                with m.Case(LinkCommand.LCRD):

                    # If the credit matches the sequence we're expecting, we can accept it!
                    with m.If(next_expected_credit == lc_detector.subtype):
                        m.d.comb += credit_received.eq(1)

                        # Next time, we'll expect the next credit in the sequence.
                        m.d.ss += next_expected_credit.eq(next_expected_credit + 1)

                    # Otherwise, we've lost synchronization. We'll need to trigger link recovery.
                    with m.Else():
                        m.d.comb += self.recovery_required.eq(1)

                #
                # Packet Acknowledgement
                #
                with m.Case(LinkCommand.LGOOD):

                    # If we've received a Header Sequence Number Advertisement, update our sequence
                    # numbers, and indicate we're done with bringup.
                    with m.If(~self.bringup_complete):
                        m.d.ss += [
                            self.bringup_complete     .eq(1),

                            next_expected_ack_number  .eq(lc_detector.subtype + 1),
                            transmit_sequence_number  .eq(lc_detector.subtype + 1)
                        ]

                    # If the credit matches the sequence we're expecting, we can accept it!
                    with m.Elif(next_expected_ack_number == lc_detector.subtype):
                        m.d.comb += retire_packet.eq(1)

                        # Next time, we'll expect the next credit in the sequence.
                        m.d.ss += next_expected_ack_number.eq(next_expected_ack_number + 1)

                    # Otherwise, if we're expecting a packet, we've lost synchronization.
                    # We'll need to trigger link recovery.
                    with m.Else():
                        m.d.comb += self.recovery_required.eq(1)

                #
                # Packet Negative Acknowledgement
                #
                with m.Case(LinkCommand.LBAD):

                    # LBADs aren't sequenced; instead, they require a retry of all unacknowledged
                    # (unretired) packets. Mark ourselves as requiring a retry; which should trigger
                    # our internal logic to execute a retry.
                    m.d.comb += self.retry_required.eq(1)


                #
                # Link Partner Retrying Send
                #
                with m.Case(LinkCommand.LRTY):

                    # If we see an LRTY, it's an indication that our link partner is performing a
                    # retried send. This information is handled by the Receiver, so we'll forward it along.
                    m.d.comb += self.retry_received.eq(1)


                #
                # Link Power State transition request.
                #
                with m.Case(LinkCommand.LGO_U):

                    # We don't handle LGO requests locally; so instead, we'll pass this back to the link layer.
                    m.d.comb += [
                        self.lgo_received  .eq(1),
                        self.lgo_target    .eq(lc_detector.subtype)
                    ]


        #
        # Header Packet Timer
        #

        # To ensure that a header packet buffer is never held by the receiver for too long
        # the USB3 specification requires us to automatically enter recovery if a credit is
        # outstanding for more than 5ms [USB3.2r1: 7.2.4.1.13]. We'll create a timer that can
        # count to this timeout.
        credit_timeout_cycles = int((self.CREDIT_TIMEOUT * self._clock_frequency + 1))
        pending_hp_timer = Signal(range(credit_timeout_cycles + 1))

        # Each time we receive a link credit and retire its packet, we'll re-start our timer.
        with m.If(retire_packet):
            m.d.ss += pending_hp_timer.eq(0)

        # If we don't have any outstanding packets, we'll clear our timer.
        with m.Elif((packets_awaiting_ack == 0) | ~self.enable):
            m.d.ss += pending_hp_timer.eq(0)

        # Otherwise, we'll count how many cycles we've had a credit outstanding for.
        with m.Else():
            m.d.ss += pending_hp_timer.eq(pending_hp_timer + 1)


        # If we ever reach our timeout, we'll need to trigger recovery.
        with m.If(pending_hp_timer == credit_timeout_cycles):
            m.d.comb += self.recovery_required.eq(1)


        #
        # Reset Handling
        #
        with m.If(~self.enable):
            m.d.ss += [
                self.bringup_complete     .eq(0),

                next_expected_credit      .eq(0),
                credits_available         .eq(0),

                packets_to_send           .eq(0),
                packets_awaiting_ack      .eq(0),
                read_pointer              .eq(0),
                write_pointer             .eq(0),
                ack_pointer               .eq(0),
                retry_pending             .eq(0),
            ]


        #
        # Debug outputs.
        #
        m.d.comb += [
            self.credits_available  .eq(credits_available),
            self.packets_to_send    .eq(packets_to_send)
        ]



        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Standard, full-gateware control request handlers. """

import os
import operator
import functools
from typing import Iterable, Callable
from warnings import warn

from luna.gateware.test import utils

from amaranth               import *
from amaranth.hdl.ast       import Value, Const
from usb_protocol.types     import USBStandardFeatures, USBStandardRequests, USBRequestRecipient, USBRequestType
from usb_protocol.emitters  import DeviceDescriptorCollection

from ..usb2.request         import RequestHandlerInterface, USBRequestHandler
from ..usb2.descriptor      import GetDescriptorHandlerDistributed, GetDescriptorHandlerBlock, GetDescriptorHandlerMux
from ..stream               import USBInStreamInterface
from ...stream.generator    import StreamSerializer
from .                      import SetupPacket
from .control               import ControlRequestHandler


class StandardRequestHandler(ControlRequestHandler):
    """ Pure-gateware USB setup request handler. Implements the standard requests required for enumeration.

    Parameters
    ----------
    descriptors: DeviceDescriptorCollection
        The DeviceDescriptorCollection that contains our descriptors.
    max_packet_size: int, optional
        The maximum packet size for the endpoint associated with this handler.
    blacklist: deprecated, use skiplist instead
    skiplist:  iterable of functions that accept a SetupPacket and return a boolean
        Collection of functions that determine if a given packet will be handled by this request handler.
    avoid_blockram: int, optional
        If True, placing data into block RAM will be avoided.

     """

    def __init__(self, descriptors: DeviceDescriptorCollection, max_packet_size=64, avoid_blockram=None, blacklist: Iterable[Callable[[SetupPacket], Value]] = (), skiplist: Iterable[Callable[[SetupPacket], Value]] = ()):
        self.descriptors      = descriptors
        self._max_packet_size = max_packet_size
        self._avoid_blockram  = avoid_blockram
        if len(blacklist) > 0:
            warn("Argument 'blacklist' is deprecated; prefer 'skiplist'.", DeprecationWarning)
            if len(skiplist) > 0:
                warn("Only one of 'blacklist' or 'skiplist' should be specified.")
            self._skiplist = blacklist
        else:
            self._skiplist = skiplist

        # If we don't have a value for avoiding blockrams; defer to the environment.
        if self._avoid_blockram is None:
            self._avoid_blockram = os.getenv("LUNA_AVOID_BLOCKRAM", False)

        super().__init__()


    def get_descriptor_handler_submodule(self):

        # The distributed handler supports a combination of fixed and runtime descriptors directly...
        if self._avoid_blockram:
            return GetDescriptorHandlerDistributed(self.descriptors, max_packet_length=self._max_packet_size)

        # ...but the block handler does not. In this case, first we split the descriptors into two
        # collections: fixed descriptors (for the ROM) and runtime descriptors.
        fixed_descriptors       = DeviceDescriptorCollection()
        runtime_descriptors     = DeviceDescriptorCollection()
        has_runtime_descriptors = False
        for type_number, index, descriptor in self.descriptors:
            if isinstance(descriptor, bytes):
                fixed_descriptors.add_descriptor(descriptor, index=index, descriptor_type=type_number)
            else:
                runtime_descriptors.add_descriptor(descriptor, index=index, descriptor_type=type_number)
                has_runtime_descriptors = True

        # If there are runtime descriptors, we add a get descriptor multiplexer and a distributed handler.
        if has_runtime_descriptors:
            handler_mux = GetDescriptorHandlerMux()
            handler_mux.add_descriptor_handler(GetDescriptorHandlerBlock(fixed_descriptors, max_packet_length=self._max_packet_size))
            handler_mux.add_descriptor_handler(GetDescriptorHandlerDistributed(runtime_descriptors, max_packet_length=self._max_packet_size))
            return handler_mux
        else:
            return GetDescriptorHandlerBlock(self.descriptors, max_packet_length=self._max_packet_size)


    def elaborate(self, platform):
        m = Module()
        interface = self.interface

        # Create convenience aliases for our interface components.
        setup               = interface.setup
        handshake_generator = interface.handshakes_out
        tx                  = interface.tx


        #
        # Submodules
        #
        # Handler for Get Descriptor requests; responds with our various fixed descriptors.
        m.submodules.get_descriptor = get_descriptor_handler = self.get_descriptor_handler_submodule()
        m.d.comb += [
            get_descriptor_handler.value  .eq(setup.value),
            get_descriptor_handler.length .eq(setup.length),
        ]

        # Handler for various small-constant-response requests (GET_CONFIGURATION, GET_STATUS).
        m.submodules.transmitter = transmitter = \
            StreamSerializer(data_length=2, domain="usb", stream_type=USBInStreamInterface, max_length_width=2)


        #
        # Handlers.
        #
        with m.If(setup.type == USBRequestType.STANDARD):

            # Only handle setup packet if not skiplisted
            skiplisted = functools.reduce(operator.__or__, (f(setup) for f in self._skiplist), Const(0))
            m.d.comb += interface.claim.eq(~skiplisted)

            with m.FSM(domain="usb"):

                # IDLE -- not handling any active request
                with m.State('IDLE'):

                    m.d.usb += [
                        # Start at the beginning of our next / fresh GET_DESCRIPTOR request.
                        get_descriptor_handler.start_position  .eq(0),

                        # Always start our responses with DATA1 pids, per [USB 2.0: 8.5.3].
                        self.interface.tx_data_pid             .eq(1)
                    ]

                    # If we've received a new setup packet, handle it.
                    with m.If(setup.received):

                        with m.If(~skiplisted):

                            # Select which standard packet we're going to handler.
                            with m.Switch(setup.request):

                                with m.Case(USBStandardRequests.GET_STATUS):
                                    m.next = 'GET_STATUS'
                                with m.Case(USBStandardRequests.CLEAR_FEATURE):
                                    m.next = 'CLEAR_FEATURE'
                                with m.Case(USBStandardRequests.SET_ADDRESS):
                                    m.next = 'SET_ADDRESS'
                                with m.Case(USBStandardRequests.SET_CONFIGURATION):
                                    m.next = 'SET_CONFIGURATION'
                                with m.Case(USBStandardRequests.GET_DESCRIPTOR):
                                    m.next = 'GET_DESCRIPTOR'
                                with m.Case(USBStandardRequests.GET_CONFIGURATION):
                                    m.next = 'GET_CONFIGURATION'
                                with m.Default():
                                    m.next = 'UNHANDLED'


                # GET_STATUS -- Fetch the device's status.
                # For now, we'll always return '0'.
                with m.State('GET_STATUS'):
                    # TODO: handle reporting endpoint stall status
                    # TODO: copy the remote wakeup and bus-powered attributes from bmAttributes of the relevant descriptor?
                    self.handle_simple_data_request(m, transmitter, 0, length=2)

                with m.State('CLEAR_FEATURE'):
                    # Provide an response to the STATUS stage.
                    with m.If(interface.status_requested):

                        # If our stall condition is met, stall; otherwise, send a ZLP [USB 8.5.3].
                        # For now, we only implement clearing ENDPOINT_HALT.
                        stall_condition = \
                            (setup.recipient != USBRequestRecipient.ENDPOINT) | \
                            (setup.value     != USBStandardFeatures.ENDPOINT_HALT)
                        with m.If(stall_condition):
                            m.d.comb += handshake_generator.stall.eq(1)
                        with m.Else():
                            m.d.comb += self.send_zlp()

                    # Accept the relevant value after the packet is ACK'd...
                    with m.If(interface.handshakes_in.ack):
                        m.d.comb += [
                            interface.clear_endpoint_halt.enable   .eq(1),
                            interface.clear_endpoint_halt.direction.eq(setup.index[7]),
                            interface.clear_endpoint_halt.number   .eq(setup.index[0:4]),
                        ]

                        # ... and then return to idle.
                        m.next = 'IDLE'

                # SET_ADDRESS -- The host is trying to assign us an address.
                with m.State('SET_ADDRESS'):
                    self.handle_register_write_request(m, interface.new_address, interface.address_changed)


                # SET_CONFIGURATION -- The host is trying to select an active configuration.
                with m.State('SET_CONFIGURATION'):
                    # TODO: stall if we don't have a relevant configuration
                    self.handle_register_write_request(m, interface.new_config, interface.config_changed)


                # GET_DESCRIPTOR -- The host is asking for a USB descriptor -- for us to "self describe".
                with m.State('GET_DESCRIPTOR'):
                    # Keep track of whether we've sent a packet we're expecting an ACK to.
                    expecting_ack = Signal()

                    m.d.comb += [
                        get_descriptor_handler.tx  .attach(tx),
                        handshake_generator.stall  .eq(get_descriptor_handler.stall)
                    ]

                    # Respond to our data stage with a descriptor...
                    with m.If(interface.data_requested):
                        m.d.comb += get_descriptor_handler.start.eq(1)
                        m.d.usb += expecting_ack.eq(1)

                    # Each time we receive an ACK, advance in our descriptor.
                    # This allows us to send descriptors with >64B of content.
                    with m.If(interface.handshakes_in.ack & expecting_ack):

                        # NOTE: this logic might need to be scaled by bytes-per-word for USB3, if it's ever used.
                        # For now, we're not using it on USB3 at all, since we assume descriptors always fit in a
                        # USB3 packet.
                        next_start_position = get_descriptor_handler.start_position + self._max_packet_size
                        m.d.usb += [

                            # We've received an ACK; so mark the section we've sent of the descriptor as
                            # received, and move forward...
                            get_descriptor_handler.start_position  .eq(next_start_position),

                            # ... and toggle our data PID.
                            self.interface.tx_data_pid             .eq(~self.interface.tx_data_pid),

                            # We've got the ACK we expected.
                            expecting_ack                          .eq(0),
                        ]

                    # ... and ACK our status stage.
                    with m.If(interface.status_requested):
                        m.d.comb += handshake_generator.ack.eq(1)
                        m.next = 'IDLE'

                    # If the requested descriptor doesn't exist, the request is terminated by STALLing the data stage.
                    with m.If(get_descriptor_handler.stall):
                        m.d.usb += expecting_ack.eq(0)
                        m.next = 'IDLE'

                # GET_CONFIGURATION -- The host is asking for the active configuration number.
                with m.State('GET_CONFIGURATION'):
                    self.handle_simple_data_request(m, transmitter, interface.active_config)


                # UNHANDLED -- we've received a request we're not prepared to handle
                with m.State('UNHANDLED'):

                    # When we next have an opportunity to stall, do so,
                    # and then return to idle.
                    with m.If(interface.data_requested | interface.status_requested):
                        m.d.comb += handshake_generator.stall.eq(1)
                        m.next = 'IDLE'

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.interface.spi import SPIGatewareTestCase
from luna.gateware.test import LunaGatewareTestCase, sync_test_case

from amaranth import Signal, Cat
from luna.gateware.debug.ila import IntegratedLogicAnalyzer, StreamILA, SyncSerialILA

class IntegratedLogicAnalyzerTest(LunaGatewareTestCase):

    def instantiate_dut(self):
        self.input_a = Signal()
        self.input_b = Signal(30)
        self.input_c = Signal()

        return IntegratedLogicAnalyzer(
            signals=[self.input_a, self.input_b, self.input_c],
            sample_depth = 32
        )


    def initialize_signals(self):
        yield self.input_a .eq(0)
        yield self.input_b .eq(0)
        yield self.input_c .eq(0)


    def provide_all_signals(self, value):
        all_signals = Cat(self.input_a, self.input_b, self.input_c)
        yield all_signals.eq(value)


    def assert_sample_value(self, address, value):
        """ Helper that asserts a ILA sample has a given value. """

        yield self.dut.captured_sample_number.eq(address)
        yield
        # Delay a clock to allow the block ram to latch the new value
        yield
        try:
            self.assertEqual((yield self.dut.captured_sample), value)
            return
        except AssertionError:
            pass

        # Generate an appropriate exception.
        actual_value = (yield self.dut.captured_sample)
        message = "assertion failed: at address 0x{:08x}: {:08x} != {:08x} (expected)".format(address, actual_value, value)
        raise AssertionError(message)


    @sync_test_case
    def test_sampling(self):

        # Quick helper that generates simple, repetitive samples.
        def sample_value(i):
            return i | (i << 8) | (i << 16) | (0xFF << 24)

        yield from self.provide_all_signals(0xDEADBEEF)
        yield

        # Before we trigger, we shouldn't be capturing any samples,
        # and we shouldn't be complete.
        self.assertEqual((yield self.dut.sampling), 0)
        self.assertEqual((yield self.dut.complete), 0)

        # Advance a bunch of cycles, and ensure we don't start sampling.
        yield from self.advance_cycles(10)
        self.assertEqual((yield self.dut.sampling), 0)

        # Set a new piece of data for a couple of cycles.
        yield from self.provide_all_signals(0x01234567)
        yield
        yield from self.provide_all_signals(0x89ABCDEF)
        yield

        # Finally, trigger the capture.
        yield from self.provide_all_signals(sample_value(0))
        yield from self.pulse(self.dut.trigger, step_after=False)

        yield from self.provide_all_signals(sample_value(1))
        yield

        # After we pulse our trigger strobe, we should be sampling.
        self.assertEqual((yield self.dut.sampling), 1)

        # Populate the memory with a variety of interesting signals;
        # and continue afterwards for a couple of cycles to make sure
        # these don't make it into our sample buffer.
        for i in range(2, 34):
            yield from self.provide_all_signals(sample_value(i))
            yield

        # We now should be done with our sampling.
        self.assertEqual((yield self.dut.sampling), 0)
        self.assertEqual((yield self.dut.complete), 1)

        # Validate the memory values that were captured.
        for i in range(32):
            yield from self.assert_sample_value(i, sample_value(i))

        # All of those reads shouldn't change our completeness.
        self.assertEqual((yield self.dut.sampling), 0)
        self.assertEqual((yield self.dut.complete), 1)


class SyncSerialReadoutILATest(SPIGatewareTestCase):

    def instantiate_dut(self):
        self.input_signal = Signal(12)
        return SyncSerialILA(
            signals=[self.input_signal],
            sample_depth=16,
            clock_polarity=1,
            clock_phase=0
        )

    def initialize_signals(self):
        yield self.input_signal.eq(0xF00)

    @sync_test_case
    def test_spi_readout(self):
        input_signal = self.input_signal

        # Trigger the test while offering our first sample.
        yield
        yield from self.pulse(self.dut.trigger, step_after=False)

        # Provide the remainder of our samples.
        for i in range(1, 16):
            yield input_signal.eq(0xF00 | i)
            yield

        # Wait a few cycles to account for delays in
        # the sampling pipeline.
        yield from self.advance_cycles(5)

        # We've now captured a full set of samples.
        # We'll test reading them out.
        self.assertEqual((yield self.dut.complete), 1)

        # Start the transaction, and exchange 16 bytes of data.
        yield self.dut.spi.cs.eq(1)
        yield

        # Read our our result over SPI...
        data = yield from self.spi_exchange_data(b"\0" * 32)

        # ... and ensure it matches what was sampled.
        i = 0
        while data:
            datum = data[0:4]
            del data[0:4]

            expected = b"\x00\x00\x0f" + bytes([i])
            self.assertEqual(datum, expected)
            i += 1

class StreamILATest(LunaGatewareTestCase):

    def instantiate_dut(self):
        self.input_signal = Signal(12)
        return StreamILA(
            signals=[self.input_signal],
            sample_depth=16
        )

    def initialize_signals(self):
        yield self.input_signal.eq(0xF00)

    @sync_test_case
    def test_stream_readout(self):
        input_signal = self.input_signal
        stream = self.dut.stream

        # Trigger the ILA with the first sample
        yield
        yield from self.pulse(self.dut.trigger, step_after=False)

        # Fill up the ILA with the remaining samples
        for i in range(1, 16):
            yield input_signal.eq(0xF00 | i)
            yield

        # Wait a few cycles to allow the ILA to fully finish processing
        yield from self.advance_cycles(6)
        # Stream should now be presenting valid data
        self.assertEqual((yield stream.valid), 1)

        # Now we want to stream out the samples from the ILA
        yield stream.ready.eq(1)
        yield
        self.assertEqual((yield stream.first), 1)

        # Read out data from the stream until it signals completion
        data = []
        while not (yield stream.last):
            if (yield stream.valid):
                data.append((yield stream.payload))
            yield

        # Match read data to what should have been sampled
        for i, datum in enumerate(data):
            self.assertEqual(datum, 0xF00 | i)



#!/usr/bin/env python3
#
# BSD 3-Clause License
#
# Copyright (c) 2018, Luke Valenty
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .           import crc
from ...usb.usb2 import USBPacketID as PID


def b(s):
    """Byte string with LSB first into an integer.

    >>> b("1")
    1
    >>> b("01")
    2
    >>> b("101")
    5
    """
    return int(s[::-1], 2)


def encode_data(data):
    """
    Converts array of 8-bit ints into string of 0s and 1s.
    """
    output = ""

    for b in data:
        output += (f"{int(b):08b}")[::-1]

    return output


def encode_pid(value):
    if not isinstance(value, PID):
        value = PID(value)
    assert isinstance(value, PID), repr(value)
    return encode_data([value.byte()])


# width=5 poly=0x05 init=0x1f refin=true refout=true xorout=0x1f check=0x19
# residue=0x06 name="CRC-5/USB"
def crc5(nibbles):
    """
    >>> hex(crc5([0, 0]))
    '0x1'
    >>> hex(crc5([3, 0]))
    '0x13'
    """
    reg = crc.CrcRegister(crc.CRC5_USB)
    for n in nibbles:
        reg.takeWord(n, 4)
    return reg.getFinalValue() & 0x1f


def crc5_token(addr, ep):
    """
    >>> hex(crc5_token(0, 0))
    '0x2'
    >>> hex(crc5_token(92, 0))
    '0x1c'
    >>> hex(crc5_token(3, 0))
    '0xa'
    >>> hex(crc5_token(56, 4))
    '0xb'
    """
    reg = crc.CrcRegister(crc.CRC5_USB)
    reg.takeWord(addr, 7)
    reg.takeWord(ep, 4)
    return reg.getFinalValue()


def crc5_sof(v):
    """
    >>> hex(crc5_sof(1429))
    '0x10'
    >>> hex(crc5_sof(1013))
    '0x14'
    """
    reg = crc.CrcRegister(crc.CRC5_USB)
    reg.takeWord(v, 11)
    return eval('0b' + bin(reg.getFinalValue() | 0x10000000)[::-1][:5])


def crc16(input_data):
    # width=16 poly=0x8005 init=0xffff refin=true refout=true xorout=0xffff
    # check=0xb4c8 residue=0xb001 name="CRC-16/USB"
    # CRC appended low byte first.
    reg = crc.CrcRegister(crc.CRC16_USB)
    for d in input_data:
        assert d <= 0xff, input_data
        reg.takeWord(d, 8)
    crc16 = reg.getFinalValue()
    return [crc16 & 0xff, (crc16 >> 8) & 0xff]


def nrzi(data, cycles=4, init="J"):
    """Converts string of 0s and 1s into NRZI encoded string.

    >>> nrzi("11 00000001", 1)
    'JJ KJKJKJKK'

    It will do bit stuffing.
    >>> nrzi("1111111111", 1)
    'JJJJJJKKKKK'

    Support single ended zero
    >>> nrzi("1111111__", 1)
    'JJJJJJKK__'

    Support pre-encoded mixing.
    >>> nrzi("11kkj11__", 1)
    'JJKKJJJ__'

    Supports wider clock widths
    >>> nrzi("101", 4)
    'JJJJKKKKKKKK'
    """
    def toggle_state(state):
        if state == 'J':
            return 'K'
        if state == 'K':
            return 'J'
        return state

    state = init
    output = ""

    stuffed = []
    i = 0
    for bit in data:
        stuffed.append(bit)
        if bit == '1':
            i += 1
        else:
            i = 0
        if i > 5:
            stuffed.append('0')
            i = 0

    for bit in stuffed:
        if bit == ' ':
            output += bit
            continue

        # only toggle the state on '0'
        if bit == '0':
            state = toggle_state(state)
        elif bit == '1':
            pass
        elif bit in "jk_":
            state = bit.upper()
        else:
            assert False, "Unknown bit %s in %r" % (bit, data)

        output += (state * cycles)

    return output


def sync():
    return "kjkjkjkk"


def eop():
    return "__j"


def wrap_packet(data, cycles=4):
    """Add the sync + eop sections and do nrzi encoding.

    >>> wrap_packet(handshake_packet(PID.ACK), cycles=1)
    'KJKJKJKKJJKJJKKK__J'
    >>> wrap_packet(token_packet(PID.SETUP, 0, 0), cycles=1)
    'KJKJKJKKKJJJKKJKJKJKJKJKJKJKKJKJ__J'
    >>> wrap_packet(data_packet(PID.DATA0, [5, 6]), cycles=1)
    'KJKJKJKKKKJKJKKKKJJKJKJKJJJKJKJKKJJJJJJKKJJJJKJK__J'
    >>> wrap_packet(data_packet(PID.DATA0, [0x1]), cycles=1)
    'KJKJKJKKKKJKJKKKKJKJKJKJJKJKJKJJJJJJJKKKJ__J'

    """
    return nrzi(sync() + data + eop(), cycles)


def token_packet(pid, addr, endp):
    """Create a token packet for testing.

    sync, pid, addr (7bit), endp(4bit), crc5(5bit), eop

    >>> token_packet(PID.SETUP, 0x0, 0x0)
    '101101000000000000001000'

     PPPPPPPP                 - 8 bits - PID
             AAAAAAA          - 7 bits - ADDR
                    EEEE      - 4 bits - EP
                        CCCCC - 5 bits - CRC

    >>> token_packet(PID.IN, 0x3, 0x0) # 0x0A
    '100101101100000000001010'

    >>> token_packet(PID.OUT, 0x3a, 0xa)
    '100001110101110010111100'

    >>> token_packet(PID.SETUP, 0x70, 0xa)
    '101101000000111010110101'

    >>> token_packet(PID.SETUP, 40, 2)
    '101101000001010010000011'

    >>> token_packet(PID.SETUP, 28, 2)
    '101101000011100010001001'

     PPPPPPPP                 - 8 bits - PID
             AAAAAAA          - 7 bits - ADDR
                    EEEE      - 4 bits - EP
                        CCCCC - 5 bits - CRC
    """
    assert addr < 128, addr
    assert endp < 2**4, endp
    assert pid in (PID.OUT, PID.IN, PID.SETUP), pid
    token = encode_pid(pid)
    token += "{0:07b}".format(addr)[::-1]  # 7 bits address
    token += "{0:04b}".format(endp)[::-1]  # 4 bits endpoint
    token += "{0:05b}".format(crc5_token(addr, endp))[::-1]  # 5 bits CRC5
    assert len(token) == 24, token
    return token


def data_packet(pid, payload):
    """Create a data packet for testing.

    sync, pid, data, crc16, eop
    FIXME: data should be multiples of 8?

    >>> data_packet(PID.DATA0, [0x80, 0x06, 0x03, 0x03, 0x09, 0x04, 0x00,\
0x02])
    '1100001100000001011000001100000011000000100100000010000000000000010000000110101011011100'

    >>> data_packet(PID.DATA1, [])
    '110100100000000000000000'

    """
    assert pid in (PID.DATA0, PID.DATA1), pid
    payload = list(payload)
    return encode_pid(pid) + encode_data(payload + crc16(payload))


def handshake_packet(pid):
    """ Create a handshake packet for testing.

    sync, pid, eop
    ack / nak / stall / nyet (high speed only)

    >>> handshake_packet(PID.ACK)
    '01001011'
    >>> handshake_packet(PID.NAK)
    '01011010'
    """
    assert pid in (PID.ACK, PID.NAK, PID.STALL), pid
    return encode_pid(pid)


def sof_packet(frame):
    """Create a SOF packet for testing.

    sync, pid, frame no (11bits), crc5(5bits), eop

    >>> sof_packet(1)
    '101001011000000000010111'

    >>> sof_packet(100)
    '101001010010011000011111'

    >>> sof_packet(257)
    '101001011000000010000011'

    >>> sof_packet(1429)
    '101001011010100110110000'

    >>> sof_packet(2**11 - 2)
    '101001010111111111111101'
    """
    def rev_byte(x):
        return int("{0:08b}".format(x)[:8][::-1], 2)

    assert frame < 2**11, (frame, '<', 2**11)
    frame_rev = int("{0:011b}".format(frame)[:11][::-1], 2)
    data = [frame_rev >> 3, (frame_rev & 0b111) << 5]
    data[-1] = data[-1] | crc5_sof(frame)
    data[0] = rev_byte(data[0])
    data[1] = rev_byte(data[1])
    return encode_pid(PID.SOF) + encode_data(data)


def diff(value):
    """Convert J/K encoding into bits for P/N diff pair.

    >>> diff('KJ_')
    ('010', '100')

    >>> # Convert ACK handshake packet
    >>> p,n = diff('KJKJKJKKJJKJJKKK__J')
    >>> p
    '0101010011011000001'
    >>> n
    '1010101100100111000'
    """
    usbp = ""
    usbn = ""
    for i in range(len(value)):
        v = value[i]
        if v == ' ':
            continue
        elif v == '_':
            # SE0 - both lines pulled low
            usbp += "0"
            usbn += "0"
        elif v == 'J':
            usbp += "1"
            usbn += "0"
        elif v == 'K':
            usbp += "0"
            usbn += "1"
        else:
            assert False, "Unknown value: %s" % v
    return usbp, usbn


def undiff(usbp, usbn):
    """Convert P/N diff pair bits into J/K encoding.

    >>> from cocotb_usb.usb.pp_packet import pp_packet
    >>> undiff(
    ...   #EJK_
    ...   '1100', # p
    ...   '1010', # n
    ... )
    'EJK_'
    >>> print(pp_packet(undiff(
    ...   #KJKJKJKKJJKJJKKK__J - ACK handshake packet
    ...   '0101010011011000001', # p
    ...   '1010101100100111000', # n
    ... ), cycles=1))
    -
    K 1 Sync
    J 2 Sync
    K 3 Sync
    J 4 Sync
    K 5 Sync
    J 6 Sync
    K 7 Sync
    K 8 Sync
    -
    J 1 PID (PID.ACK)
    J 2 PID
    K 3 PID
    J 4 PID
    J 5 PID
    K 6 PID
    K 7 PID
    K 8 PID
    -
    _ SE0
    _ SE0
    J END
    >>> print(pp_packet(undiff(*diff(wrap_packet(sof_packet(0))))))
    ----
    KKKK 1 Sync
    JJJJ 2 Sync
    KKKK 3 Sync
    JJJJ 4 Sync
    KKKK 5 Sync
    JJJJ 6 Sync
    KKKK 7 Sync
    KKKK 8 Sync
    ----
    KKKK 1 PID (PID.SOF)
    JJJJ 2 PID
    JJJJ 3 PID
    KKKK 4 PID
    JJJJ 5 PID
    JJJJ 6 PID
    KKKK 7 PID
    KKKK 8 PID
    ----
    JJJJ  1 Frame #
    KKKK  2 Frame #
    JJJJ  3 Frame #
    KKKK  4 Frame #
    JJJJ  5 Frame #
    KKKK  6 Frame #
    JJJJ  7 Frame #
    KKKK  8 Frame #
    ----
    JJJJ  9 Frame #
    KKKK 10 Frame #
    JJJJ 11 Frame #
    KKKK 1 CRC5
    KKKK 2 CRC5
    JJJJ 3 CRC5
    KKKK 4 CRC5
    JJJJ 5 CRC5
    ----
    ____ SE0
    ____ SE0
    JJJJ END
    """
    assert len(usbp) == len(
        usbn), "Sequence different lengths!\n%s\n%s\n" % (usbp, usbn)
    value = []
    for i in range(0, len(usbp)):
        p = usbp[i]
        n = usbn[i]
        value.append({
            # pn
            '00': '_',
            '11': 'E',
            '10': 'J',
            '01': 'K',
        }[p + n])
    return "".join(value)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
